## 
## Copyright 1999, 2000, 2001, 2003, 2004, 2005, 2006, 2007, 2008, 2009, 2011, 2012, 2019 The Regents of the University of California
## All Rights Reserved 
## 
## Permission to use, copy, modify and distribute any part of this
## CoralReef software package for educational, research and non-profit
## purposes, without fee, and without a written agreement is hereby
## granted, provided that the above copyright notice, this paragraph
## and the following paragraphs appear in all copies.
## 
## Those desiring to incorporate this into commercial products or use
## for commercial purposes should contact the Technology Transfer
## Office, University of California, San Diego, 9500 Gilman Drive, La
## Jolla, CA 92093-0910, Ph: (858) 534-5815, FAX: (858) 534-7345.
## 
## IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY
## PARTY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL
## DAMAGES, INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS
## SOFTWARE, EVEN IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF
## THE POSSIBILITY OF SUCH DAMAGE.
## 
## THE SOFTWARE PROVIDED HEREIN IS ON AN "AS IS" BASIS, AND THE
## UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE,
## SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS. THE UNIVERSITY
## OF CALIFORNIA MAKES NO REPRESENTATIONS AND EXTENDS NO WARRANTIES
## OF ANY KIND, EITHER IMPLIED OR EXPRESS, INCLUDING, BUT NOT LIMITED
## TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR A
## PARTICULAR PURPOSE, OR THAT THE USE OF THE SOFTWARE WILL NOT INFRINGE
## ANY PATENT, TRADEMARK OR OTHER RIGHTS.
## 
## The CoralReef software package is developed by the CoralReef
## development team at the University of California, San Diego under
## the Cooperative Association for Internet Data Analysis (CAIDA)
## Program. Support for this effort is provided by the CAIDA grant
## NCR-9711092, DARPA NGI Contract N66001-98-2-8922, DARPA NMS Grant
## N66001-01-1-8909, and by CAIDA members.
## 
## Report bugs and suggestions to coral-bugs@caida.org.
## 

##########################################################################
#
# Makefile.inc.in
#
# This file is helpful for writing a Makefile for a project that uses
# CoralReef libraries.  To do so, put the following at the beginning of
# the project's Makefile:
#   CORAL_DIR=/usr/local/Coral  # change as needed
# and this at the end:
#   include $(CORAL_DIR)/lib/Makefile.vars
#   include $(CORAL_DIR)/lib/Makefile.inc
#
# You may define any of the following special variables in the Makefile.  All
# are optional unless otherwise noted, but to be useful at least one of
# C_PROGS, CXX_PROGS, or PL_SRCS should be defined, and the default "all"
# target should depend on "depend" and on $(C_PROGS), $(CXX_PROGS), and/or
# $(PL_PROGS).  The "depend" file must exist before you can use make
# ("touch depend" is sufficient).
# Note: the project Makefile should _not_ set CC or MAKE.
#
# CORAL_DIR	Installed CoralReef directory (REQUIRED).
# C_PROGS	List of programs; each will be built from a single .o file
#		with the same base name.  For programs built from multiple
#		.o files, you will have to define the rule yourself.
# CXX_PROGS	List of programs; each will be built from a single .o file
#		with the same base name.  For programs built from multiple
#		.o files, you will have to define the rule yourself.
# PL_SRCS	List of perl sources (with a ".pl" suffix).  Each will be
#		built into a suffixless file.
# C_SRCS	List of .c files; each will be used to create a dependency
#		rule for a corresponding .o file.
# CXX_SRCS	List of .cc files; each will be used to create a dependency
#		rule for a corresponding .o file.
# EXTRA_LIBS	-L and -l options to be used when linking C programs.
# EXTRA_LDFLAGS	Flags to be used when linking C programs.
# EXTRA_CPPFLAGS Flags to be used when preprocessing C code, e.g.,
#		-I$(CORAL_SRCDIR)
# EXTRA_CFLAGS	Flags to be used when compiling or linking C code.
# EXTRA_CXXFLAGS	Flags to be used when compiling or linking C++ code.
# C_PROG_DEPS	Additional dependencies of C_PROGS, e.g.,
#		$(CORAL_SRCDIR)/lib/libcoral.a
# CXX_PROG_DEPS	Additional dependencies of CXX_PROGS, e.g.,
#		$(CORAL_SRCDIR)/lib/libcoral.a
# PL_PROG_DEPS	Additional dependencies of $(PL_SRCS:.pl=), e.g.,
#		$(CORAL_SRCDIR)/Makefile.vars
#
#
# The following variables are set in this Makefile.inc, and may be used
# by Makefiles that are built with it:
#
# CC		C compiler
# CFLAGS	flags for CC, not including preprocessor flags
# CPP		C preprocessor
# CPPFLAGS	flags for CPP
# RANLIB	ranlib
# SWIG_CC	C compiler needed by SWIG
# CXX		C++ compiler
# CXXFLAGS	C++ compiler flags, not including preprocessor flags
# PERL		perl 5 interpreter
# PL_PROGS	list of programs to be built from PL_SRCS
# SWIG		swig
# RRD_DIR	RRDtool directory
# FIXPERL	fixperl (replaces #! line and specified variables)
# FIXPERLFLAGS	fixperl flags
# LIBS		-l options for linker, common to C and C++
# CC_LIBS	C-specific libraries (not including LIBS)
# CXX_LIBS	C++-specific libraries (not including LIBS)
# LDFLAGS	linker options common to C and C++
# CC_LDFLAGS	all linker options for C (including LDFLAGS)
# CXX_LDFLAGS	all linker options for C++ (including LDFLAGS)
#
##########################################################################

LD = ld
AR = ar

.SUFFIXES: .pl .html .pod .obj .pdf .gif .pm .so .swig-c .swig-cc
.PRECIOUS: depend
.PHONY: all install install_doc clean realclean dist

# Usage:  @$(OPTIONAL) "command args..."
# If CORAL_OPTIONAL is unset (the usual case), an error in command will not
# be fatal, it will only print a warning message.  If CORAL_OPTIONAL is "no",
# an error in command will be fatal.
OPTIONAL= \
	optional() { \
	    echo "$$*"; \
	    if test "$${CORAL_OPTIONAL}" = "no"; then \
		eval "$$*"; \
	    else \
		eval "$$* || echo 'WARNING: unable to build optional $@'"; \
	    fi; \
	}; \
	optional

# We generate depend at distribution time, so we must exclude any
# system-dependent files:
# - Swig-generated code includes perl headers
# - any other file with an absolute path
# We use perl because sed is too inconsistent from system to system.
FIXDEP= perl -pe 's% /\S*\b%%g'

# We assume at least one of mkdep or gcc is available anywhere coral
# development is done.
MKDEP=fn_mkdep() { gcc -MM "$$@" >depend.tmp || mkdep -f depend.tmp "$$@"; }; fn_mkdep

# Notes:
# Automatic inclusion of ".depend" is a pmake feature, not portable.
# "ifneq" etc is a gmake feature, not portable.
# If "depend" was created by configure, its timestamp is incorrect, so we make
# it depend on ".depend.time".
# We use egrep '[^ ]' instead of test -n because vars may contain spaces but
# no filenames.
depend: .depend.time
	@echo "Updating '`pwd`/depend'."
	@echo "# Automatically generated by 'make depend'." >depend
	@echo "" >>depend
	@if echo "$(C_OBJS)" | egrep '[^ ]' >/dev/null; then \
	    echo "    C objects: $(C_OBJS)"; \
	    $(MKDEP) $(CFLAGS) $(CPPFLAGS) $(C_SRCS) && \
		$(FIXDEP) depend.tmp>>depend || \
	        { rm .depend.time; false; } \
	fi
	@list="$(C_PROGS)"; \
	if echo "$(C_PROGS)" | egrep '[^ ]' >/dev/null; then \
	    echo "    C programs: $(C_PROGS)"; \
	    for f in $${list:-dummy}; do \
	        echo "$$f: $$f.o $(C_PROGDEPS)" >>depend; \
		echo '	$$(CC_LD) $$(CFLAGS) -o' $$f $$f.o '$$(C_COMMON_OBJS)' '$$(CC_LDFLAGS)' >>depend; \
	    done; \
	fi
	@if echo "$(CXX_OBJS)" | egrep '[^ ]' >/dev/null; then \
	    echo "    C++ objects: $(CXX_OBJS)"; \
	    $(MKDEP) $(CXXFLAGS) $(CPPFLAGS) $(CXX_SRCS) && \
		$(FIXDEP) depend.tmp>>depend || \
	        { rm .depend.time; false; } \
	fi
	@list="$(CXX_PROGS)"; \
	if echo "$(CXX_PROGS)" | egrep '[^ ]' >/dev/null; then \
	    echo "    C++ programs: $(CXX_PROGS)"; \
	    for f in $${list:-dummy}; do \
	        echo "$$f: $$f.o $(CXX_PROGDEPS)" >>depend; \
		echo '	$$(CXX_LD) $$(CXXFLAGS) -o' $$f $$f.o '$$(CXX_COMMON_OBJS)' '$$(CXX_LDFLAGS)' >>depend; \
	    done; \
	fi
	@list="$(PL_PROGS)"; \
	if echo "$${list}" | egrep '[^ ]' >/dev/null; then \
	    echo "    perl programs: $${list}"; \
	    for f in $${list}; do \
	        echo "$$f: $$f.pl $(PL_PROGDEPS)" >>depend; \
		echo '	$$(FIXPERL) $$(FIXPERLFLAGS)' $$f.pl >>depend; \
	    done; \
	fi
	@-rm depend.tmp

.depend.time: $(C_SRCS) $(CXX_SRCS) $(PL_SRCS)
	@touch .depend.time


# A target that other targets can depend on to force them to always be remade
# (.PHONY is a pmake feature, not portable)
FORCE:

# Implicit rules
.c.o:
	$(CC) -c $(CFLAGS) $(CPPFLAGS) $<

.cc.o:
	$(CXX) -c $(CXXFLAGS) $(CPPFLAGS) $<

# bug: this is not right for C++
.o:
	$(CC_LD) $(CFLAGS) -o $@ $*.o $(CC_LDFLAGS)

.c:
	$(CC) $(CFLAGS) $(CPPFLAGS) -o $@ $*.c $(CC_LDFLAGS)

.pl:
	$(FIXPERL) $(FIXPERLFLAGS) $*.pl

.pl.html:
	@$(OPTIONAL) "pod2html $(P2HFLAGS) $< --outfile=$*.html"

.pm.html:
	@$(OPTIONAL) "pod2html $(P2HFLAGS) $< --outfile=$*.html"

.pod.html:
	@$(OPTIONAL) "pod2html $(P2HFLAGS) $< --outfile=$*.html"


# Rules to build images from tgif .obj files.  Uses $(OPTIONAL) so errors
# are fatal for "make dist", but not for normal user builds.
# To work around a tgif bug, we set TMPDIR if it isn't already. (A leading
# VAR=value won't work with OPTIONAL, must use env).
.obj.gif:
	@rm -f $@
	@$(OPTIONAL) "env TMPDIR=$${TMPDIR:-/tmp} tgif -print -gif $<"

.obj.pdf:
	@rm -f $@
	@$(OPTIONAL) "env TMPDIR=$${TMPDIR:-/tmp} tgif -print -pdf $<"

include depend
