/* 
 * Copyright 2004, 2007, 2008, 2009, 2011, 2012, 2019 The Regents of the University of California
 * All Rights Reserved 
 * 
 * Permission to use, copy, modify and distribute any part of this
 * CoralReef software package for educational, research and non-profit
 * purposes, without fee, and without a written agreement is hereby
 * granted, provided that the above copyright notice, this paragraph
 * and the following paragraphs appear in all copies.
 * 
 * Those desiring to incorporate this into commercial products or use
 * for commercial purposes should contact the Technology Transfer
 * Office, University of California, San Diego, 9500 Gilman Drive, La
 * Jolla, CA 92093-0910, Ph: (858) 534-5815, FAX: (858) 534-7345.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY
 * PARTY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL
 * DAMAGES, INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE SOFTWARE PROVIDED HEREIN IS ON AN "AS IS" BASIS, AND THE
 * UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE,
 * SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS. THE UNIVERSITY
 * OF CALIFORNIA MAKES NO REPRESENTATIONS AND EXTENDS NO WARRANTIES
 * OF ANY KIND, EITHER IMPLIED OR EXPRESS, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR A
 * PARTICULAR PURPOSE, OR THAT THE USE OF THE SOFTWARE WILL NOT INFRINGE
 * ANY PATENT, TRADEMARK OR OTHER RIGHTS.
 * 
 * The CoralReef software package is developed by the CoralReef
 * development team at the University of California, San Diego under
 * the Cooperative Association for Internet Data Analysis (CAIDA)
 * Program. Support for this effort is provided by the CAIDA grant
 * NCR-9711092, DARPA NGI Contract N66001-98-2-8922, DARPA NMS Grant
 * N66001-01-1-8909, and by CAIDA members.
 * 
 * Report bugs and suggestions to coral-bugs@caida.org.
 */
/*
 * $Id: crl_anf.cc,v 1.68 2019/03/28 23:40:30 kkeys Exp $
 */
/*
 * NetFlow-like system with adaptive sampling probability
 */
/* XXX TODO: do something useful with -p prefixlen option for IPv6?
 */

static const char RCSid[]="$Id: crl_anf.cc,v 1.68 2019/03/28 23:40:30 kkeys Exp $";

#include "config.h"
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <cstddef>
#include <math.h>
#include <string.h>
#include <ctype.h>
#include <sys/time.h>
#include <time.h>
#include <sys/errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <signal.h>
#include <errno.h>
#include <stdexcept>

#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <netinet/tcp.h>
#include <netinet/udp.h>
#include <netinet/ip_icmp.h>
#ifdef HAVE_INET6
# include <netinet/ip6.h>
#endif
#include <arpa/inet.h>
#include <netdb.h>
#include <sys/resource.h> // getrusage
#include <sys/wait.h> /* for waitpid */

#include "libcoral.h"
#include "crl_byteorder.h"
#include "hashtab.h"

#include <utility>
#if defined(HAVE_UNORDERED_MAP) // C++0x standard
# include <unordered_map>
# include <unordered_set>
#elif defined(HAVE_TR1_UNORDERED_MAP) // g++ 4, maybe others
# include <tr1/unordered_map>
# include <tr1/unordered_set>
#elif defined(HAVE_HASH_MAP) // g++ 3, SGI...
# include <hash_map>
# include <hash_set>
#elif defined(HAVE_EXT_HASH_MAP) // g++ 3, ...
# include <ext/hash_map>
# include <ext/hash_set>
#else
# error "Cannot build without unordered_map and unordered_set"
#endif

#undef NDEBUG
#include <assert.h>

#include "h3.h"
#include "poolallocator.h"
#include "alignmentof.h"

//#define OUTPUT_VERSION "1.1"
#define OUTPUT_VERSION "1.2"	    // fixed renormalization of flow's syn flag

// Hack: key bits and counter bits are mixed into a single byte on output
// so existing CoralReef table readers didn't need to be modified to read
// a new counter column.
#define PORTS_NONE   0x00 // port fields are not valid
#define PORTS_LEGACY 0x01 // dport and sport have valid values (old format ONLY)
#define PORTS_DPORT  0x02 // dport is a valid dst port
#define PORTS_SPORT  0x04 // sport is a valid src port
#define PORTS_PORTS  (PORTS_SPORT | PORTS_DPORT)
#define PORTS_OTHER  0x08 // "ports" are really other proto-dependent values
			  // (icmp: type/code)

#if !PATH_MAX
# define PATH_MAX 1024
#endif

using namespace std;
using namespace HASH_NAMESPACE;


class SampProb {
public:
    uint32_t d; // denominator

    SampProb(uint32_t pd = 1) : d(pd) {};

    operator double() { return 1.0 / double(d); }
    // don't define conversion to integer, to avoid accidents

    bool operator== (const SampProb &b) const { return d == b.d; }
    SampProb& operator*= (const SampProb &b) { d *= b.d; return *this; }
    SampProb& operator/= (uint32_t &i) { d *= i; return *this; }
};

inline SampProb operator* (const SampProb &a, const SampProb &b) {
    SampProb r = a;
    return r *= b;
}

inline SampProb operator/ (const SampProb &a, uint32_t i) {
    SampProb r = a;
    return r /= i;
}



typedef uint64_t tuple_hashval_t;
#define PRIxTUP PRIx64

typedef uint64_t flow_t;

static const char *hostname(const struct in_addr *addr);
#ifdef HAVE_INET6
static const char *hostname(const struct in6_addr *addr);
#endif

class Procstats
{
    const char *firstheap;
    const char *lastheap;
    struct timeval firstcpu, lastcpu;
public:
    Procstats() {
	firstheap = lastheap = (char*)sbrk(0);
	lastcpu.tv_sec = 0;
	lastcpu.tv_usec = 0;
    }
    ~Procstats() {
#ifndef NDEBUG
	(*this)("end app");
#endif
    }
    void operator()(const char *label) {
	struct timeval cpu, cpudiff;
	if (coral_verbosity < 3) return;
	struct rusage ru;
	const char *thisheap = (char*)sbrk(0);
	getrusage(RUSAGE_SELF, &ru);
	timeradd(&ru.ru_utime, &ru.ru_stime, &cpu);
	coral_printf("# %s: heap %ld KB, cpu %ld.%03lds ",
	    label, (thisheap - firstheap) / 1024, cpu.tv_sec, cpu.tv_usec/1000);
	timersub(&cpu, &lastcpu, &cpudiff);
	coral_printf("(+%ld KB, +%ld.%03lds)\n",
	    (thisheap - lastheap) / 1024, cpudiff.tv_sec, cpudiff.tv_usec/1000);
	lastheap = thisheap;
	lastcpu = cpu;
    }
    void mem(const struct timespec *ts);
} procstats;

struct config {
    double dinterval;
    double E;
    u_int netmask;
    uint8_t allow_v4;
    uint8_t allow_v6;
    SampProb pkt_prob;
    SampProb byte_prob;
    SampProb flow_prob;
    u_char anf;
    u_char fce;
    size_t anf_report_size;
    size_t fce_report_size;
    size_t anf_normal_size;
    size_t anf_max_size;
    u_char mem;
    int normalizer_pkts;
    int normalizer_entries;
//    u_char sampling; // XXX not currently used --rkoga
    u_char syn_counting;
    u_char normalize_immediately;
    u_char unscaled;
    u_char cnames;
    u_char stats;
    u_char fork;
    u_char binary;
    u_char pretty;  /* non-pretty isn't needed if apps can read pretty format */
    u_char rotate_files;
//    u_char skip_ephemeral_ports; // XXX not currently used --rkoga
//    u_char est_prec; // XXX not currently used --rkoga
    u_char merge_ifaces;
    const char *outfilename;
    char *file_header;
    coral_rotfile_t *rf;
    FILE *file;
    coral_anon_t *anon;
} config;

struct interval_data {
    double ip4_bytes;
    double ip6_bytes;
    double notIP_pkts;
    coral_pkt_stats_t *pkt_stats;
    double begin;
    double end;
} interval_data;

#define VPVC_HASH_TABLE_SIZE	229

// assuming random() generates at least [0, 2^31-1], we mask it off to
// exactly [0, 2^31-1]
const uint32_t RAND_MASK = 0x7FFFFFFFul;
const uint32_t RAND_RANGE = 0x80000000ul;

const unsigned PKT_BKTS = 32;

struct subif_stats;

template<class A>
struct tuple_key {
    A src, dst;
    uint16_t sport;
    uint16_t dport;
    uint8_t proto;
    uint8_t ports_ok;
    static const uint8_t length;
    static const char *fmt_lheader, *fmt_rheader;
    static const char *fmt_addr, *fmt_right;

    bool operator== (const tuple_key &b) const
    {
	return
	    sport == b.sport && dport == b.dport &&
	    proto == b.proto && ports_ok == b.ports_ok &&
	    memcmp(&src, &b.src, sizeof(A)) == 0 &&
	    memcmp(&dst, &b.dst, sizeof(A)) == 0;
    }

    static void header()
    {
	fprintf(config.file, fmt_lheader, "src     ", "dst     ");
	fprintf(config.file, fmt_rheader, "pro", "ok", "sport", "dport");
    }

    void dump_text() const
    {
	fprintf(config.file, fmt_addr, hostname(&src));
	fprintf(config.file, fmt_addr, hostname(&dst));
	fprintf(config.file, fmt_right,
	    proto, ports_ok, sport, dport);
    }

    void dump_binary() const
    {
        uint8_t full_save = 0; // there are no first and latest timestamps
        uint16_t nsport = htons(sport);
        uint16_t ndport = htons(dport);
        fwrite(&length, sizeof(length), 1, config.file);
        fwrite(&full_save, sizeof(full_save), 1, config.file);
        fwrite(&src, sizeof(src), 1, config.file);
        fwrite(&dst, sizeof(dst), 1, config.file);
        fwrite(&proto, sizeof(proto), 1, config.file);
        fwrite(&ports_ok, sizeof(ports_ok), 1, config.file);
        fwrite(&nsport, sizeof(nsport), 1, config.file);
        fwrite(&ndport, sizeof(ndport), 1, config.file);
    }
};

template<class A> const uint8_t tuple_key<A>::length =
    2 * sizeof(A) + 2 * sizeof(uint16_t) + 2 * sizeof(uint8_t);

template<class Key, class T>
class tuple_key_hash {
    const H3<T, Key::length> h;
public:
    tuple_key_hash() : h() {};

    T operator()(const Key &key) const
    {
	return h(&key, Key::length, 0);
    }
};

BEGIN_HASH_NAMESPACE
template<>
struct hash<tuple_key<struct in_addr> > {
    size_t operator()(const tuple_key<struct in_addr> & key) const
    {
	return /* (size_t)(key.parent) ^ */ // XXX must include if/subif
	    ((size_t)(key.src.s_addr) * 59) ^
	    ((size_t)(key.dst.s_addr)) ^
	    ((size_t)(key.sport) << 16) ^
	    ((size_t)(key.dport)) ^
	    ((size_t)(key.proto));
    }
};

#ifdef HAVE_INET6
template<>
struct hash<tuple_key<struct in6_addr> > {
    size_t operator()(const tuple_key<struct in6_addr> & key) const
    {
	return /* (size_t)(key.parent) ^ */ // XXX must include if/subif
	    ((size_t)((static_cast<const uint32_t*>((void*)&key.src))[3]) * 59) ^
	    ((size_t)((static_cast<const uint32_t*>((void*)&key.dst))[3])) ^
	    ((size_t)(key.sport) << 16) ^
	    ((size_t)(key.dport)) ^
	    ((size_t)(key.proto));
    }
};
#endif
END_HASH_NAMESPACE

static const char *fmt_dataheader, *fmt_data;
template<class A> const char *tuple_key<A>::fmt_lheader;
template<class A> const char *tuple_key<A>::fmt_rheader;
template<class A> const char *tuple_key<A>::fmt_addr;
template<class A> const char *tuple_key<A>::fmt_right;


// calculates floor(log2(x)) fairly efficiently
inline int flog2(uint32_t x) {
    int r = 0;
#define TEST(m,e) if (x & m) {x &= m; r += e;} else {x &= ~m;}
    TEST(0xFFFF0000ul, 16)
    TEST(0xFF00FF00ul, 8)
    TEST(0xF0F0F0F0ul, 4)
    TEST(0xCCCCCCCCul, 2)
    TEST(0xAAAAAAAAul, 1)
#undef TEST
    return r;
}

// calculates floor(log2(x)) fairly efficiently
inline int flog2(uint64_t x) {
    int r = 0;
#define TEST(m,e) if (x & m) {x &= m; r += e;} else {x &= ~m;}
    TEST(0xFFFFFFFF00000000ull, 32)
    TEST(0xFFFF0000FFFF0000ull, 16)
    TEST(0xFF00FF00FF00FF00ull, 8)
    TEST(0xF0F0F0F0F0F0F0F0ull, 4)
    TEST(0xCCCCCCCCCCCCCCCCull, 2)
    TEST(0xAAAAAAAAAAAAAAAAull, 1)
#undef TEST
    return r;
}

// ceil(log2(x))
inline uint32_t clog2(uint32_t x) { return flog2(x - 1) + 1; }
inline uint64_t clog2(uint64_t x) { return flog2(x - 1) + 1; }

struct flow_data; // forward declaration

// unordered_map plus sampling probabilities for packet counting synopsis
template <class Key, class HashFcn = hash<Key>, class EqualKey = equal_to<Key> >
class ANFFlowtab : public unordered_map<Key, flow_data*, HashFcn, EqualKey>
{
    typedef unordered_map<Key, flow_data*, HashFcn, EqualKey> HM;
    ANFFlowtab operator=(const ANFFlowtab& x)
    {
	coral_diag(1, ("ANFFlowtab error, copy constructor not implemented\n"));
	abort();
//	throw NotImplemented;  // Where is NotImplemented defined?
    }
    SampProb _pkt_prob, _byte_prob;
    uint32_t _psampn;

    void init()
    {
	_psampn = random() % _pkt_prob.d; // random starting offset
	for (unsigned i = 0; i < PKT_BKTS; i++) pbkt[i].hist = 0;
    }

/* XXX Not currently used. --rkoga
    struct scanner : public unary_function<HM::value_type&, void> {
	uint32_t one, many;
	void operator() (HM::value_type &val) {
	    flow_data *data = val.second;
	    // assert(data->pkt_prob == _pkt_prob);
	    if (data->pkts == 1) one++;
	    else if (data->pkts > 1) many++;
	}
    };
*/

    struct {		// information about entries with a given number of pkts
	uint64_t hist;	// histogram: number of entries with this many pkts
	uint64_t fmean;	// normalizer: floor of mean of distribution
	uint64_t total;	// normalizer: # of entries tested 
	uint64_t inc;	// normalizer: # of entries incremented
	uint32_t rem;	// normalizer: remainder
    } pbkt[PKT_BKTS];

    bool normalizing;
    int normalizer_pkts;
    int normalizer_tested;
    int normalizer_zeroed;
    typename HM::iterator normalizer_it;
    typename HM::iterator normalizer_end;

public:
    ANFFlowtab(SampProb ppkt_prob, SampProb pbyte_prob) :
	HM(), _pkt_prob(ppkt_prob), _byte_prob(pbyte_prob),
	_psampn(), pbkt(), normalizing(false)
	{ init(); }
    ANFFlowtab(SampProb ppkt_prob, SampProb pbyte_prob,
	size_t n) :
	HM(n), _pkt_prob(ppkt_prob), _byte_prob(pbyte_prob),
	_psampn(), pbkt(), normalizing(false)
	{ init(); }
    ANFFlowtab(SampProb ppkt_prob, SampProb pbyte_prob,
	size_t n, const typename HM::hasher& h) :
	HM(n,h), _pkt_prob(ppkt_prob), _byte_prob(pbyte_prob),
	_psampn(), pbkt(), normalizing(false)
	{ init(); }
    ANFFlowtab(SampProb ppkt_prob, SampProb pbyte_prob,
	size_t n, const typename HM::hasher& h,
	const typename HM::key_equal& k) :
	HM(n,h,k), _pkt_prob(ppkt_prob), _byte_prob(pbyte_prob),
	_psampn(), pbkt(), normalizing(false)
	{ init(); }
    ANFFlowtab(SampProb ppkt_prob, SampProb pbyte_prob,
	const ANFFlowtab& x) :
	HM(x), _pkt_prob(ppkt_prob), _byte_prob(pbyte_prob),
	_psampn(), pbkt(), normalizing(false)
	{ init(); }

    SampProb pkt_prob() { return _pkt_prob; }
    SampProb byte_prob() { return _byte_prob; }
    SampProb pkt_prob(SampProb prob)
	{ _pkt_prob = prob; return _pkt_prob; }
    SampProb byte_prob(SampProb prob)
	{ return _byte_prob = prob; /* XXX */ }
    bool should_samplepkt()
	{ if (++_psampn < _pkt_prob.d) return false; _psampn = 0; return true; }
    bool should_samplebyte()
	{ return false; /* XXX */ }
    bool resample(flow_data *data);
    void samplepkt(const Key &key, uint64_t subif_pkts, uint32_t bytes, bool syn);

    int sizetobucket(uint64_t n) { return n < PKT_BKTS ? n : 0; }

    void adj_pkt_hist(uint64_t n0, uint64_t n1)
    {
	if (n0 > 0) pbkt[sizetobucket(n0)].hist--;
	if (n1 > 0) pbkt[sizetobucket(n1)].hist++;
    }

    uint64_t pkt_est()
    {
	uint64_t pkts = 0;
	for (typename HM::iterator it = this->begin(); it != this->end(); it++){
	    pkts += it->second->pkts;
	}
	return pkts * _pkt_prob.d;
    }

    uint64_t byte_est()
    {
	uint64_t bytes = 0;
	for (typename HM::iterator it = this->begin(); it != this->end(); it++){
	    bytes += it->second->bytes;
	}
	return bytes * _pkt_prob.d;
    }

    void start_normalizer(bool force, uint64_t true_pkts)
    {
	if ((force && this->size() > config.anf_report_size) ||
	    (!normalizing && this->size() >= config.anf_normal_size) ||
	    (this->size() >= config.anf_max_size))
	{
	    uint64_t t0;    // reciprocal of old sampling prob
	    uint64_t t1;    // reciprocal of new sampling prob
	    unsigned b;	    // bucket number
	    uint64_t goal;  // # of entries we want to delete
	    uint64_t E;	    // # of entries we expect to delete with current t1
	    uint64_t sum;   // running sum of (pbkt[b].hist * (t1 - b*t0));
	    uint64_t cdf;   // CDF of pbkt[1].hist ... pbkt[b-1].hist

	    // pick new pkt_prob
	    {
		if (config.normalize_immediately) { // XXX debugging
		    static uint64_t old_true_pkts = 0;
		    coral_diag(3, ("starting normalizer: packets: %" PRIu64
			" (+%" PRIu64 ")\n",
			true_pkts, true_pkts - old_true_pkts));
		    old_true_pkts = true_pkts;
		}

#ifndef NDEBUG
		procstats("starting normalizer");
#endif
		goal = this->size() - config.anf_report_size;
		// assert(_pkt_prob.n == 1);
		t0 = _pkt_prob.d;

		// quickly pick reasonable starting point for t1
		t1 = (pbkt[1].hist < goal * 2) ? t0 * 2 + 1 : t0 + 1;

		// calculate cdf and sum for initial set of buckets
		sum = 0;
		cdf = 0;
		for (b = 1; t0 * b < t1; b++) {
		    sum += (pbkt[b].hist * (t1 - b*t0));
		    cdf += pbkt[b].hist;
		}

		// while we haven't reached goal, update cdf and sum
		uint64_t max_t1 = PKT_BKTS * t0;
		while (sum < goal * t1) {
		    sum += cdf;
		    t1++;
		    assert(t1 < max_t1); // XXX must handle this
		    // t1++ may include one additional bucket
		    if (t0 * b < t1) {
			sum += pbkt[b].hist * (t1 - b*t0);
			cdf += pbkt[b].hist;
			b++;
		    }
		}

		E = uint64_t(sum / double(t1));
		// _pkt_prob.n = 1;
		_pkt_prob.d = t1;
	    }

	    if (normalizing) {
		coral_diag(3, ("normalizer REstarted: size %d, prob %u/%" PRIu32 ", "
		    "tested %d, zeroed %d\n", this->size(), 1, _pkt_prob.d,
		    normalizer_tested, normalizer_zeroed)); // XXX
	    } else {
		coral_diag(3, ("normalizer started: size %d, prob %u/%" PRIu32 "\n",
		    this->size(), 1, _pkt_prob.d)); // XXX
	    }
	    if (coral_verbosity >= 3) {
		coral_diag(3, ("normalizer goal %" PRIu64 ", expect %" PRIu64 "\n",
		    goal, E)); // XXX
		coral_diag(3, ("normalizer pkt hist: "));
		for (b = 1; b < PKT_BKTS; b++)
		    coral_diag(3, (" %" PRIu64, pbkt[b].hist));
		coral_diag(3, (" (%" PRIu64 ")\n", pbkt[0].hist));
	    }

	    // start normalizer
	    normalizing = true;
	    normalizer_it = this->begin();
	    normalizer_pkts = config.normalizer_pkts;
	    normalizer_zeroed = 0;
	    normalizer_tested = 0;

	    for (b = 1; b < PKT_BKTS; b++) {
		// NB: integer division
		pbkt[b].fmean = b * t0 / t1;
		pbkt[b].rem = b * t0 % t1;

		// start with random offset into cycle to avoid bias
		pbkt[b].total = random() % t1;
		pbkt[b].inc = pbkt[b].total * pbkt[b].rem / t1;

		coral_diag(3, ("normalizer state: bkt %2d, entries %6" PRIu64
		    ", mean %3" PRIu64 " +%4" PRIu32 "/%" PRIu32
		    " (%4" PRIu64 "/%" PRIu64 ")\n",
		    b, pbkt[b].hist,
		    pbkt[b].fmean, pbkt[b].rem, _pkt_prob.d,
		    pbkt[b].inc, pbkt[b].total)); // XXX
	    }
	}
    }

    void run_normalizer(bool to_end);

    inline void delete_contents()
    {
	typename HM::iterator it;
	for (it = this->begin(); it != this->end(); ++it)
	    delete it->second;
	this->clear();
    }
};

// unordered_set plus sampling probability
template <class Key, class HashFcn = hash<Key>, class EqualKey = equal_to<Key> >
class FCEFlowtab : public unordered_set<Key, HashFcn, EqualKey>
{
    typedef unordered_set<Key, HashFcn, EqualKey> HS;
    static tuple_key_hash<Key, tuple_hashval_t> *tuple_hash;
    FCEFlowtab operator=(const FCEFlowtab& x)
    {
	coral_diag(1, ("FCEFlowtab error, copy constructor not implemented\n"));
	abort();
//	throw NotImplemented;  // Where is NotImplemented defined?
    }
    SampProb _flow_prob;
    tuple_hashval_t _fsampthresh;

    void init()
    {
	_fsampthresh = tuple_hashval_t(~(tuple_hashval_t)0 / _flow_prob.d);
    }

    void clean_flows()
    {
	typename HS::iterator flowit, next;
	for (flowit = this->begin(); flowit != this->end(); flowit = next)
	{
	    next = flowit;
	    ++next;
	    if (!should_sampleflow(*flowit)) {
		this->erase(flowit);
	    }
 	}
    }

public:
    static void static_init() {
	tuple_hash = new tuple_key_hash<Key, tuple_hashval_t>();
    }
    FCEFlowtab(SampProb pflow_prob) :
	HS(), _flow_prob(pflow_prob)
	{ init(); }
    FCEFlowtab(SampProb pflow_prob, size_t n) :
	HS(n), _flow_prob(pflow_prob)
	{ init(); }
    FCEFlowtab(SampProb pflow_prob, size_t n, const typename HS::hasher& h) :
	HS(n,h), _flow_prob(pflow_prob)
	{ init(); }
    FCEFlowtab(SampProb pflow_prob, size_t n, const typename HS::hasher& h,
	const typename HS::key_equal& k) :
	HS(n,h,k), _flow_prob(pflow_prob)
	{ init(); }
    FCEFlowtab(SampProb pflow_prob, const FCEFlowtab& x) :
	HS(x), _flow_prob(pflow_prob)
	{ init(); }

    SampProb flow_prob() { return _flow_prob; }
    SampProb flow_prob(SampProb prob)
	{ _flow_prob = prob; init(); return _flow_prob; }
    bool should_sampleflow(const Key &key)
	{ return (*tuple_hash)(key) < _fsampthresh; }

    uint64_t flow_est()
    {
	return (uint64_t)(this->size() * _flow_prob.d);
    }

    void flow_normalizer(bool final)
    {
	if (this->size() > (final ? 1.1 : 2.0) * config.fce_report_size) {
	    coral_diag(3, ("cleaner start: size %d, prob %u/%" PRIu32 ", thresh %"
		PRIxTUP "\n", this->size(), 1, _flow_prob.d, _fsampthresh)); // XXX

	    if (final) {
		double factor = double(config.fce_report_size) / this->size();
		_flow_prob.d = uint32_t(_flow_prob.d / factor);
		// We can't just multiply _fsampthresh by factor, because
		// _flow_prob.d was rounded down to an integer.  We must
		// recalculate _fsampthesh based on the rounded _flow_prob.d.
		_fsampthresh = tuple_hashval_t(~(tuple_hashval_t)0 /
				    _flow_prob.d);
	    } else {
		_flow_prob.d *= 2;
		_fsampthresh = _fsampthresh >> 1;
	    }

	    clean_flows();

	    coral_diag(3, ("cleaner done: size %d, prob %u/%" PRIu32 ", thresh %"
		PRIxTUP "\n", this->size(), 1, _flow_prob.d, _fsampthresh)); // XXX
	}
    }
};

template <class Key, class HashFcn, class EqualKey>
tuple_key_hash<Key, tuple_hashval_t> *FCEFlowtab<Key, HashFcn, EqualKey>::tuple_hash;

typedef uint64_t pktcnt;
typedef uint64_t bytecnt;

struct flow_data {
    pktcnt pkts;	// must be 8-byte aligned on some architectures
    bytecnt bytes;	// must be 8-byte aligned on some architectures
    bool syn;
    SampProb pkt_prob;	// table's pkt_prob when pkt counter was last updated

    flow_data() : pkts(), bytes(), syn(false), pkt_prob() {}
    ~flow_data() {}
};

// The obviously correct thing to do would be to effectively resample
// every packet by calculating a new count with a binomial distribution
// in (pkts, p); but that's expensive.  Instead, we find the (real) mean
// of the binomial distribution, and pick the (integer) floor or ceiling
// of the mean with probability such that the resulting distribution has
// the same mean.  This is still unbiased, and actually improves variance.
template <class Key, class HashFcn, class EqualKey>
bool ANFFlowtab<Key, HashFcn, EqualKey>::resample(flow_data *data)
{
    bool inc;
    pktcnt n = data->pkts;

    this->normalizer_tested++;
    if (n == 0) {
	return false; // shouldn't happen
    } else if (n < PKT_BKTS) {
	// use precalculated mean and increment-ratio
	this->pbkt[n].total++;
	data->pkts = this->pbkt[n].fmean;
	inc = (this->pbkt[n].inc + 1) * this->pkt_prob().d <=
	    this->pbkt[n].total * this->pbkt[n].rem;
	if (inc) {
	    this->pbkt[n].inc++;
	    data->pkts++;
	}
    } else {
	// calculate mean and random increment
	uint32_t t0 = data->pkt_prob.d;
	uint32_t t1 = this->pkt_prob().d;
	data->pkts = n * t0 / t1;
	inc = (random() % t1) < (n * t0 % t1); // (n*t0 % t1) in t1 odds
	if (inc)
	    data->pkts++;
    }
    if (config.syn_counting && data->syn) {
	// odds of keeping syn = # of pkts kept / original # of pkts
	for (int i = data->syn; i > 0; i--) {
	    if ((random() % n) >= data->pkts)
		data->syn = false;
	}
    }

    data->bytes = data->bytes * data->pkts / n;
    data->pkt_prob = this->pkt_prob();
    this->adj_pkt_hist(n, data->pkts);
    if (data->pkts == 0 && data->bytes == 0) {
	this->normalizer_zeroed++;
	return true;
    }
    return false;
}

template <class Key, class HashFcn, class EqualKey>
void ANFFlowtab<Key, HashFcn, EqualKey>::run_normalizer(bool to_end)
{
    if (normalizing && (to_end || --normalizer_pkts == 0)) {
	int entries = config.normalizer_entries;
	typename HM::iterator flowit;

	for (flowit = normalizer_it; flowit != this->end();
	    flowit = normalizer_it)
	{
	    ++normalizer_it;
	    flow_data *data = flowit->second;
	    if (data->pkt_prob != pkt_prob()) {
		if (this->resample(data)) {
		    delete data;
		    this->erase(flowit);
		}
	    }
	    if (!to_end && --entries == 0) break;
	}

	if (flowit == this->end()) {
	    normalizing = false;
	    coral_diag(3, ("normalizer done: size %d, prob %u/%" PRIu32 ", "
		"tested %d, zeroed %d\n", this->size(), 1, _pkt_prob.d,
		normalizer_tested, normalizer_zeroed)); // XXX
#ifndef NDEBUG
	    procstats("normalizer done");
#endif
	} else {
	    normalizer_pkts = config.normalizer_pkts;
	}
    }
}

template<class A>
struct flowset_t {
    static const int v;
    pktcnt pkts;
    bytecnt bytes;
    ANFFlowtab<tuple_key<A> > *anf_tuple_tab;
    FCEFlowtab<tuple_key<A> > *fce_tuple_tab;

    flowset_t() : pkts(), bytes(), anf_tuple_tab(), fce_tuple_tab() {}
    inline void init() {
	if (!this->anf_tuple_tab && (config.anf || config.syn_counting)) {
	    this->anf_tuple_tab = new ANFFlowtab<tuple_key<A> >(config.pkt_prob,
		config.byte_prob, config.anf_report_size);
	}
	if (!this->fce_tuple_tab && config.fce) {
	    this->fce_tuple_tab = new FCEFlowtab<tuple_key<A> >(config.flow_prob,
		config.fce_report_size);
	}
    }
    void clear() {
	if (anf_tuple_tab) {
	    anf_tuple_tab->delete_contents();
	    delete anf_tuple_tab;
	    anf_tuple_tab = NULL;
	}
	if (fce_tuple_tab) {
	    delete fce_tuple_tab;
	    fce_tuple_tab = NULL;
	}
    }

    void dump_header() const {
	if (this->pkts > 0) {
	    fprintf(config.file, "#      IPv%d pkts: %" PRIu64 "\n",
		this->v, this->pkts);
	    if (this->anf_tuple_tab) {
		fprintf(config.file, "#  est IPv%d pkts: %" PRIu64 "\n",
		    this->v, this->anf_tuple_tab->pkt_est());
	    }
	    fprintf(config.file, "#     IPv%d bytes: %" PRIu64 "\n",
		this->v, this->bytes);
	    if (this->anf_tuple_tab) {
		fprintf(config.file, "# est IPv%d bytes: %" PRIu64 "\n",
		    this->v, this->anf_tuple_tab->byte_est());
	    }
	    fprintf(config.file, "#     IPv%d flows: 0\n", this->v);
	    if (this->fce_tuple_tab) {
		fprintf(config.file, "# est IPv%d flows: %" PRIu64 "\n",
		    this->v, this->fce_tuple_tab->flow_est());
	    }
	    fprintf(config.file,
		"#     IPv%d first: 0\n"
		"#    IPv%d latest: 0\n",
		this->v, this->v); // XXX
	}
    }
    void dump_binary_flow_table() const;
    void dump_text_flow_table() const;
    void dump_flow_table(const subif_stats *subifrec, const char *label) const;
};

template<> const int flowset_t<in_addr>::v = 4;
#ifdef HAVE_INET6
template<> const int flowset_t<in6_addr>::v = 6;
#endif

struct subif_stats {
// key
    coral_iface_t *iface;
    u_int iface_id;
    u_int subif_id;

// data
    pktcnt unknown_encaps;
    flowset_t<in_addr> ip4;
#ifdef HAVE_INET6
    flowset_t<in6_addr> ip6;
#endif

    subif_stats(coral_iface_t *p_iface, u_int p_iface_id, u_int p_subif_id) :
	iface(p_iface), iface_id(p_iface_id), subif_id(p_subif_id),
	unknown_encaps(), ip4()
#ifdef HAVE_INET6
	, ip6()
#endif
	{}

    void clear() {
	ip4.clear();
#ifdef HAVE_INET6
	ip6.clear();
#endif
    }

    ~subif_stats() { clear(); }
};

hash_tab *subif_hash = NULL;

/* XXX kkeys: needs to be generalized to multiple subifs */
void Procstats::mem(const struct timespec *ts) {
    static const struct subif_stats *subifrec = NULL; // XXX hack, assumes only one subif
    if (!subifrec) {
	init_hash_walk(subif_hash);
	subifrec = (const subif_stats*)next_hash_walk(subif_hash);
    }
    const char *thisheap = (char*)sbrk(0);
    coral_printf("# %ld.%06ld\t%ld\t%ld\t%ld\t%ld\t%ld\n",
	ts->tv_sec, ts->tv_nsec/1000,
	(thisheap - firstheap) / 1024,
	subifrec->ip4.anf_tuple_tab ? subifrec->ip4.anf_tuple_tab->size() : 0,
	subifrec->ip4.fce_tuple_tab ? subifrec->ip4.fce_tuple_tab->size() : 0,
#ifdef HAVE_INET6
	subifrec->ip6.anf_tuple_tab ? subifrec->ip6.anf_tuple_tab->size() : 0,
	subifrec->ip6.fce_tuple_tab ? subifrec->ip6.fce_tuple_tab->size() : 0
#else
	0, 0
#endif
	);
}

volatile int interrupted = 0;
const char rotating_filename[] = "%010s.%f.t2";


/* ---- hash table per subif fxns begin ---- */
/* -- per subif hash table fxns -- */

/* return 0 if the same - for use by the hashtable */
static int compare_subif_stats(const void *entry1, const void *entry2)
{
    const subif_stats *foo1 = (const subif_stats *)entry1;
    const subif_stats *foo2 = (const subif_stats *)entry2;

    return (foo1->subif_id != foo2->subif_id ||
	    foo1->iface_id != foo2->iface_id);
}

/* make a hash of an entry - for use by the hashtable */
static unsigned long make_key_subif_stats(const void *entry)
{
    const subif_stats *what = (const subif_stats *)entry;

    return (unsigned) what->subif_id * 425879 + what->iface_id;
}

/* free mem of an entry - for use by the hashtable */
static void delete_subif_stats(void *entry)
{
    subif_stats *what = (subif_stats *)entry;
    if (!what) return;
    free(what);
}
/* -- end of subif type hash table fxns -- */


/* ---- hash table fxns end ---- */

template <class Key, class HashFcn, class EqualKey>
void ANFFlowtab<Key, HashFcn, EqualKey>::samplepkt(
    const Key &key, uint64_t subif_pkts, uint32_t bytes, bool syn)
{
    if (this->should_samplepkt()) {
	flow_data *data = 0;
	typename ANFFlowtab::iterator flowit;
	flowit = (this->insert(
	    typename ANFFlowtab::value_type(key, NULL))).first;
	if (flowit->second) {
	    // found existing entry
	    data = flowit->second;
#if 1
	    // about to update an entry that needs renormalizing
	    if (data->pkt_prob != this->pkt_prob()) {
		this->resample(data);
	    }
#endif
	} else {
	    // inserted new entry
	    flowit->second = data = new flow_data();
	    this->start_normalizer(false, subif_pkts);
	    if (config.normalize_immediately)
		this->run_normalizer(true);
	}
	pktcnt oldpkts = data->pkts++;
	data->pkt_prob = this->pkt_prob();
	this->adj_pkt_hist(oldpkts, data->pkts);
	data->bytes += bytes;
	if (syn) data->syn = true;
    }
    this->run_normalizer(false);
}


static void count_packet(subif_stats *subif, coral_iface_t *iface,
    const coral_pkt_buffer_t *netpkt, const coral_timestamp_t *ts)
{
    short offset = 0;
    uint32_t bytes = 0;
    coral_pkt_buffer_t payload;
    uint16_t sport = 0, dport = 0;
    int ports_ok = 0;
    bool syn = false;
    uint8_t proto = 0;

    tuple_key<in_addr> flowid4;
#ifdef HAVE_INET6
    tuple_key<in6_addr> flowid6;
#endif

    if (netpkt->protocol == CORAL_NETPROTO_IPv4) {
	struct ip* ip4 = (struct ip*)netpkt->buf;
	offset = ntohs(ip4->ip_off) & IP_OFFMASK;
	flowid4.src.s_addr = config.netmask & ip4->ip_src.s_addr;
	flowid4.dst.s_addr = config.netmask & ip4->ip_dst.s_addr;
	bytes = ntohs(ip4->ip_len);
	interval_data.ip4_bytes += bytes;
	subif->ip4.pkts++;
	subif->ip4.bytes += bytes;
#ifdef HAVE_INET6
    } else if (netpkt->protocol == CORAL_NETPROTO_IPv6) {
	struct ip6_hdr* ip6 = (struct ip6_hdr*)netpkt->buf;
	flowid6.src = ip6->ip6_src;
	flowid6.dst = ip6->ip6_dst;
	bytes = sizeof(*ip6) + ntohs(ip6->ip6_plen);
	interval_data.ip6_bytes += bytes;
	subif->ip6.pkts++;
	subif->ip6.bytes += bytes;
#endif
    }

    coral_get_payload(netpkt, &payload);
    /* NOT get_payload_by_layer(..., 4); we want the payload, no matter what
     * layer it claims to be.  (Specifically, we don't want to go through
     * IPIP encapsulations.)
     */

    proto = coralproto_to_ipproto(payload.protocol);
    /* We can't simply look at the IP header: there may be IPv6 ext hdrs. */

    /* extract port info */
    if (offset != 0) {
	// non-first fragment, no layer 4 header
    } else if (proto == IPPROTO_TCP) {
	if (payload.caplen >= 4) {
	    struct tcphdr *t = (struct tcphdr*)payload.buf;
	    sport = ntohs(t->th_sport);
	    dport = ntohs(t->th_dport);
	    ports_ok = PORTS_PORTS;
	    syn = !!(t->th_flags & TH_SYN);
	}
    } else if (proto == IPPROTO_UDP) {
	if (payload.caplen >= 4) {
	    struct udphdr *u = (struct udphdr*)payload.buf;
	    sport = ntohs(u->uh_sport);
	    dport = ntohs(u->uh_dport);
	    ports_ok = PORTS_PORTS;
	}
    } else if (proto == IPPROTO_ICMP) {
	if (payload.caplen >= 2) {
	    struct icmp *icmp = (struct icmp*)payload.buf;
	    sport = icmp->icmp_type;
	    dport = icmp->icmp_code;
	    ports_ok = PORTS_OTHER;
	}
#ifdef HAVE_INET6
    } else if (proto == IPPROTO_ICMPV6) {
	if (payload.caplen >= 2) {
	    struct icmp *icmp = (struct icmp*)payload.buf;
	    sport = icmp->icmp_type;
	    dport = icmp->icmp_code;
	    ports_ok = PORTS_OTHER;
	}
#endif
#if 0
    } else if (proto == IPPROTO_IGMP) {
	    ...
	    sport = igmp->igmp_type;
	    dport = igmp->igmp_code;
	    ports_ok = PORTS_OTHER;
#endif
    }

    if (netpkt->protocol == CORAL_NETPROTO_IPv4) {
	flowid4.proto = proto;
	flowid4.ports_ok = ports_ok;
	flowid4.sport = sport;
	flowid4.dport = dport;
	if (subif->ip4.fce_tuple_tab && subif->ip4.fce_tuple_tab->should_sampleflow(flowid4)) {
	    subif->ip4.fce_tuple_tab->insert(flowid4);
	    subif->ip4.fce_tuple_tab->flow_normalizer(false);
	}
	if (subif->ip4.anf_tuple_tab) {
	    subif->ip4.anf_tuple_tab->samplepkt(flowid4, subif->ip4.pkts, bytes, syn);
	}

#ifdef HAVE_INET6
    } else if (netpkt->protocol == CORAL_NETPROTO_IPv6) {
	flowid6.proto = proto;
	flowid6.ports_ok = ports_ok;
	flowid6.sport = sport;
	flowid6.dport = dport;
	if (subif->ip6.fce_tuple_tab && subif->ip6.fce_tuple_tab->should_sampleflow(flowid6)) {
	    subif->ip6.fce_tuple_tab->insert(flowid6);
	    subif->ip6.fce_tuple_tab->flow_normalizer(false);
	}
	if (subif->ip6.anf_tuple_tab) {
	    subif->ip6.anf_tuple_tab->samplepkt(flowid6, subif->ip6.pkts, bytes, syn);
	}
#endif
    }
}

/* NB: returns a pointer to static data */
static const char *hostname(const struct in_addr *addr)
{
    struct hostent *h;
    if (config.anon) {
	struct in_addr newaddr;
	coral_anonymize_ip4(config.anon, &newaddr, addr);
	return inet_ntoa(newaddr);
    }
    if (config.cnames &&
	(h = gethostbyaddr((char*)addr, sizeof(struct in_addr), AF_INET)))
    {
	return h->h_name;
    } else {
	return inet_ntoa(*addr);
    }
}

#ifdef HAVE_INET6
/* NB: returns a pointer to static data */
static const char *hostname(const struct in6_addr *addr)
{
    struct hostent *h;
    static char buf[64];
    if (config.anon) {
	struct in6_addr newaddr;
	coral_anonymize_ip6(config.anon, &newaddr, addr);
	return inet_ntop(AF_INET6, &newaddr, buf, sizeof(buf));
    }
    if (config.cnames &&
	(h = gethostbyaddr((char*)addr, sizeof(struct in6_addr), AF_INET6)))
    {
        return h->h_name;
    } else {
	return inet_ntop(AF_INET6, (char*)addr, buf, sizeof(buf));
    }
}
#endif

template<class A>
void flowset_t<A>::dump_binary_flow_table() const
{
    ANFFlowtab<tuple_key<A> > *anf = this->anf_tuple_tab;
    FCEFlowtab<tuple_key<A> > *fce = this->fce_tuple_tab;
    fputs("#binary data follows\n", config.file);

    if (anf) {
	uint32_t factor = config.unscaled ? 1 : anf->pkt_prob().d;
	typename ANFFlowtab<tuple_key<A> >::iterator i;
	for (i = anf->begin(); i != anf->end(); ++i) {
	    if (i->second->pkts) {
		i->first.dump_binary();
		uint64_t flow_pkts = i->second->pkts * factor;
		uint64_t flow_bytes = i->second->bytes * factor;
		uint64_t flow_flows = config.syn_counting && i->second->syn ?
		    factor : 0;
		flow_pkts = crl_hton64(flow_pkts);
		flow_bytes = crl_hton64(flow_bytes);
		flow_flows = crl_hton64(flow_flows);
		fwrite(&flow_pkts, sizeof(flow_pkts), 1, config.file);
		fwrite(&flow_bytes, sizeof(flow_bytes), 1, config.file);
		fwrite(&flow_flows, sizeof(flow_flows), 1, config.file);
	    }
	}
    }

    if (fce) {
	uint32_t factor = config.unscaled ? 1 : fce->flow_prob().d;
	typename FCEFlowtab<tuple_key<A> >::iterator i;
	for (i = fce->begin(); i != fce->end(); ++i) {
	    i->dump_binary();
	    uint64_t flow_pkts = 0;
	    uint64_t flow_bytes = 0;
	    uint64_t flow_flows = factor;
	    flow_pkts = crl_hton64(flow_pkts);
	    flow_bytes = crl_hton64(flow_bytes);
	    flow_flows = crl_hton64(flow_flows);
	    fwrite(&flow_pkts, sizeof(flow_pkts), 1, config.file);
	    fwrite(&flow_bytes, sizeof(flow_bytes), 1, config.file);
	    fwrite(&flow_flows, sizeof(flow_flows), 1, config.file);
	}
    }

    fwrite("\0", 1, 1, config.file); // table trailer
    fputs("\n# end of binary table\n\n", config.file);

    fflush(config.file);
}

template<class A>
void flowset_t<A>::dump_text_flow_table() const
{
    ANFFlowtab<tuple_key<A> > *anf = this->anf_tuple_tab;
    FCEFlowtab<tuple_key<A> > *fce = this->fce_tuple_tab;
    uint64_t total_pkts = 0, total_bytes = 0, total_flows = 0;

    tuple_key<A>::header();
    fprintf(config.file, fmt_dataheader, "pkts", "bytes", "flows");

    if (anf) {
	uint32_t factor = config.unscaled ? 1 : anf->pkt_prob().d;
	typename ANFFlowtab<tuple_key<A> >::iterator i;
	for (i = anf->begin(); i != anf->end(); ++i) {
	    i->first.dump_text();
	    unsigned int flows = config.syn_counting && i->second->syn;
	    total_pkts += i->second->pkts;
	    total_bytes += i->second->bytes;
	    fprintf(config.file, fmt_data, i->second->pkts * factor,
		i->second->bytes * factor, flows * factor);
	    total_flows += flows;
	}
    }
    if (fce) {
	uint32_t factor = config.unscaled ? 1 : fce->flow_prob().d;
	typename FCEFlowtab<tuple_key<A> >::iterator i;
	for (i = fce->begin(); i != fce->end(); ++i) {
	    i->dump_text();
	    fprintf(config.file, fmt_data, uint64_t(0), uint64_t(0),
		uint32_t(factor));
	    total_flows++;
	}
    }
    fprintf(config.file, "# pkts=%" PRIu64 " bytes=%" PRIu64,
	total_pkts, total_bytes);
    fprintf(config.file, " flows=%" PRIu64 "\n", total_flows);
    fputs("# end of text table\n\n", config.file);

    fflush(config.file);
}

template<class A>
void flowset_t<A>::dump_flow_table(const subif_stats *subifrec,
    const char *label) const
{
    // if (table.empty()) return;
    char buf[32] = "0";
    if (!config.merge_ifaces)
	coral_fmt_if_subif(buf, subifrec->iface, subifrec->subif_id);

    // before "# begin" line to avoid choking parser
    // tuple_key::notes();
    fprintf(config.file, "# Counter values are%s already scaled.\n",
	config.unscaled ? " NOT" : "");
    if (config.anf) {
	fprintf(config.file, "# initial pkt sample prob: %u/%" PRIu32 "\n",
	    1, config.pkt_prob.d);
#if 0
	fprintf(config.file, "# initial byte sample prob: %u/%" PRIu32 "\n",
	    1, config.byte_prob.d);
#endif
    }
    if (config.fce) {
	fprintf(config.file, "# initial flow sample prob: %u/%" PRIu32 "\n",
	    1, config.flow_prob.d);
    }
    if (config.anf) {
	fprintf(config.file, "# final pkt sample prob: %u/%" PRIu32 "\n",
	    1, this->anf_tuple_tab->pkt_prob().d);
#if 0
	fprintf(config.file, "# final byte sample prob: %u/%" PRIu32 "\n",
	    1, table.byte_prob().d);
#endif
    }
    if (config.fce) {
	fprintf(config.file, "# final flow sample prob: %u/%" PRIu32 "\n",
	    1, this->fce_tuple_tab->flow_prob().d);
    }
    if (config.syn_counting) {
	fprintf(config.file, "# TCP SYN in flow column\n");
    }
    fprintf(config.file, "# begin %s Table ID: %s (%zd entries)\n", label, buf,
	(this->anf_tuple_tab ? this->anf_tuple_tab->size() : 0) +
	(this->fce_tuple_tab ? this->fce_tuple_tab->size() : 0));
    if (config.binary) {
	this->dump_binary_flow_table();
    } else {
	this->dump_text_flow_table();
    }
}

/* 
 * keep track of some subif related stats, and count_packet().
 */
static void count_subif_stats(coral_iface_t *iface,
    coral_pkt_result_t *pkt_result)
{
    subif_stats *subifrec;
    subif_stats tmp(iface, coral_interface_get_number(iface), pkt_result->subiface);
    coral_pkt_buffer_t netpkt;

    if (config.merge_ifaces) {
	tmp.iface = NULL;
	tmp.iface_id = 0;
	tmp.subif_id = 0;
    }

    if (!(subifrec = (subif_stats*)find_hash_entry(subif_hash, &tmp))) {
	subifrec = new subif_stats(tmp.iface, tmp.iface_id, tmp.subif_id);
	if (subifrec == NULL) {
	    fprintf(stderr, "can't allocate subif_stats.\n");
	    abort();
	}
        add_hash_entry(subif_hash, subifrec);
#ifndef NDEBUG
	procstats("init hash");
#endif
    }

    if (coral_get_payload_by_layer(pkt_result->packet, &netpkt, 3) < 0) {
	subifrec->unknown_encaps++;
	return;
    }

    if (netpkt.protocol == CORAL_NETPROTO_IPv4) {
	if (!config.allow_v4) return;
	subifrec->ip4.init();
    } else if (netpkt.protocol == CORAL_NETPROTO_IPv6) {
	if (!config.allow_v6) return;
#ifdef HAVE_INET6
	subifrec->ip6.init();
#endif
    } else {
	interval_data.notIP_pkts++;
	return;
    }

    /* do stats --- */
    count_packet(subifrec, iface, &netpkt, pkt_result->timestamp);
}

static void init_hashes(void)
{
    subif_hash = init_hash_table("# subif hash",
			       compare_subif_stats, make_key_subif_stats,
			       delete_subif_stats, VPVC_HASH_TABLE_SIZE);
    if (config.fce) {
	if (config.allow_v4)
	    FCEFlowtab<tuple_key<in_addr> >::static_init();
#ifdef HAVE_INET6
	if (config.allow_v6)
	    FCEFlowtab<tuple_key<in6_addr> >::static_init();
#endif
    }
}


static void dump_hash_stats(void)
{
    dump_hashtab_stats(subif_hash);
}

static void dump_data(void)
{
    const subif_stats *subifrec;
    double begin;
    double end;
    int count;

    begin = interval_data.begin;
    end = interval_data.end;

    /* Normal (possibly empty) interval */
    fprintf(config.file, "\n# begin trace interval: %f\n", begin);
    fprintf(config.file, "# trace interval duration: %f s\n"
		    "# Layer 2 PDUs dropped: %" PRIu64 "\n"
		    "# Packets dropped: %" PRIu64 "\n",
	(end - begin), 
	interval_data.pkt_stats->l2_drop,
	interval_data.pkt_stats->pkts_drop);
    if (config.allow_v4)
	fprintf(config.file, "# IPv4: %.4f Mbit/s\n",
	    (double)interval_data.ip4_bytes * 8 / (1000 * 1000 * (end - begin)));
    if (config.allow_v6)
	fprintf(config.file, "# IPv6: %.4f Mbit/s\n",
	    (double)interval_data.ip6_bytes * 8 / (1000 * 1000 * (end - begin)));
    fprintf(config.file, "# Non-IP: %.4f pkts/s\n",
	interval_data.notIP_pkts/(end - begin));

    fputs("# Table IDs: ", config.file);
    init_hash_walk(subif_hash);
    count = 0;
    while ((subifrec = (const subif_stats*)next_hash_walk(subif_hash))) {
	if (subifrec->ip4.anf_tuple_tab || subifrec->ip4.anf_tuple_tab ||
	    subifrec->ip4.fce_tuple_tab || subifrec->ip4.fce_tuple_tab ||
	    subifrec->unknown_encaps)
	{
	    char buf[32] = "0";
	    if (count++)
		fputs(", ", config.file);
	    if (!config.merge_ifaces)
		coral_fmt_if_subif(buf, subifrec->iface, subifrec->subif_id);
	    fprintf(config.file, "%s", buf);
	}
    }
    fputs("\n\n", config.file);

    init_hash_walk(subif_hash);
    while ((subifrec = (const subif_stats*)next_hash_walk(subif_hash))) {
	if (subifrec->ip4.anf_tuple_tab || subifrec->ip4.fce_tuple_tab ||
#ifdef HAVE_INET6
	    subifrec->ip6.anf_tuple_tab || subifrec->ip6.fce_tuple_tab ||
#endif
	    subifrec->unknown_encaps)
	{
	    if (subifrec->ip4.anf_tuple_tab) {
		// finish ANF normalization in progress, if any
		subifrec->ip4.anf_tuple_tab->run_normalizer(true);
		// normalize to get the requested number of ANF entries
		subifrec->ip4.anf_tuple_tab->start_normalizer(true, subifrec->ip4.pkts);
		subifrec->ip4.anf_tuple_tab->run_normalizer(true);
	    }
	    if (subifrec->ip4.fce_tuple_tab) {
		// XXX resample to EXACTLY the right size (not half)
		subifrec->ip4.fce_tuple_tab->flow_normalizer(true);
	    }
#ifdef HAVE_INET6
	    if (subifrec->ip6.anf_tuple_tab) {
		// finish ANF normalization in progress, if any
		subifrec->ip6.anf_tuple_tab->run_normalizer(true);
		// normalize to get the requested number of ANF entries
		subifrec->ip6.anf_tuple_tab->start_normalizer(true, subifrec->ip6.pkts);
		subifrec->ip6.anf_tuple_tab->run_normalizer(true);
	    }
	    if (subifrec->ip6.fce_tuple_tab) {
		// XXX resample to EXACTLY the right size (not half)
		subifrec->ip6.fce_tuple_tab->flow_normalizer(true);
	    }
#endif

	    char buf[32] = "0";
	    if (!config.merge_ifaces)
		coral_fmt_if_subif(buf, subifrec->iface, subifrec->subif_id);
	    fprintf(config.file, "# ID: %s\n"
		    "# unknown_encaps: %" PRIu64 "\n",
		    buf,
		    subifrec->unknown_encaps);
	    if (config.allow_v4) subifrec->ip4.dump_header();
#ifdef HAVE_INET6
	    if (config.allow_v6) subifrec->ip6.dump_header();
#endif
	    int types = 0;
	    fprintf(config.file, "# Table types: ");
	    if (config.allow_v4 && subifrec->ip4.pkts) {
		if (types) fputs(", ", config.file);
		fprintf(config.file, "Tuple Table");
		types++;
	    }
#ifdef HAVE_INET6
	    if (config.allow_v6 && subifrec->ip6.pkts) {
		if (types) fputs(", ", config.file);
		fprintf(config.file, "IPv6 Tuple Table");
		types++;
	    }
#endif
	    fputs("\n\n", config.file); 
	}
    }

    init_hash_walk(subif_hash);
    while ((subifrec = (const subif_stats*)next_hash_walk(subif_hash))) {
	if (config.allow_v4 && subifrec->ip4.pkts)
	    subifrec->ip4.dump_flow_table(subifrec, "Tuple");
#ifdef HAVE_INET6
	if (config.allow_v6 && subifrec->ip6.pkts)
	    subifrec->ip6.dump_flow_table(subifrec, "IPv6 Tuple");
#endif
    }
    fputs("# end trace interval\n", config.file);
    fflush(config.file);
}

static void reap_children(int options) {
    pid_t pid = 0;
    int status = 0;

    if (config.fork) {
	while ((pid = waitpid(-1, &status, options)) > 0) {
	    // XXX should probably check the status
	    coral_diag(2, ("crl_anf: Reaping child process %d: %d\n",
		pid, status));
	    status = 0;
	}
    }
}

static void dump_data_fork(struct timeval *tvp)
{
    int inside_child = 0;
    struct timeval start_tv;
    static int first = 1;

    reap_children(WNOHANG);

    if (config.rotate_files || first) {
	coral_rf_start_name(config.rf, tvp);
    }

    if (config.fork) {
	pid_t pid;

	gettimeofday(&start_tv, NULL);

	pid = fork();
	if (pid > 0) {
	    // fork succeeded, this is parent
	    coral_diag(2, ("crl_anf: [%d] Dumping using child process\n", pid));
	    return;
	} else if (pid < 0) {
	    // Fork failed
	    coral_diag(0, ("crl_anf: Unable to fork child dumper process: %s\n",
		strerror(errno)));
	} else {
	    inside_child = 1;
	}
    }

    if (config.rotate_files || first) {
	coral_rf_start_output(config.rf);
	fputs(config.file_header, config.file);
	first = 0;
    }

    dump_data();

    if (config.rotate_files) {
	coral_rf_end(config.rf);
    }

    if (inside_child) {
	pid_t pid;
	struct timeval end_tv, delta_tv;

	pid = getpid();
	gettimeofday(&end_tv, NULL);
	timersub(&end_tv, &start_tv, &delta_tv);
	coral_diag(2, ("crl_anf: [%d] Child dumper process took %d.%06d s\n",
		       pid, delta_tv.tv_sec, delta_tv.tv_usec));

	/* NB: at fork time, child may have inherited other partially filled
	 * stdio buffers.  We use _exit() instead of exit() to avoid flushing
	 * them. */
	_exit(0);
    }
}

static void clear_data(void)
{
    subif_stats *subifrec;

    interval_data.ip4_bytes = 0;
    interval_data.ip6_bytes = 0;
    interval_data.notIP_pkts = 0;

    init_hash_walk(subif_hash);
    while ((subifrec = (subif_stats*)next_hash_walk(subif_hash))) {
	subifrec->clear();
    }
    clear_hash_table(subif_hash);
}

static void stop(int sig)
{
    interrupted = 1;
}

int main(int argc, char *argv[])
{
    int retval = 0;
    int i;
    coral_iface_t *iface;
    coral_pkt_result_t pkt_result;
    coral_interval_result_t interval_result;
#   define CIDR_LEN 32
    u_int cidr_len = CIDR_LEN;
    struct timeval interval = { 300, 0 };
    u_int duration;
    int opt;
    int error = 0, seeded = 0;
    char file_header[1024], *p;

    config.outfilename = "-"; /* Default to stdout */

    coral_set_api(CORAL_API_PKT);
    coral_set_duration(0);
    coral_set_interval(&interval);
    coral_set_iomode(0, CORAL_RX, 48, 0);
    coral_set_options(0, CORAL_OPT_PARTIAL_PKT | CORAL_OPT_NORMALIZE_TIME);

    config.anf = 1;
    config.fce = 0;
    config.anf_report_size = 0;
    config.fce_report_size = 0;
    config.syn_counting = 0;
    config.normalizer_pkts = 8;
    config.normalizer_entries = 24;
    config.unscaled = 0;
    config.fork = 0;

    const char *options = "C:p:asbf:yhO:o:rS:eA:K:Y:F:mnuR46";

    while (!error && (opt = getopt(argc, argv, options)) != -1) {
	switch (opt) {
	case 'C':
	    if (coral_config_command(optarg) < 0)
		error++;
	    break;
	case 'p':
	    cidr_len = atoi(optarg);
	    if (cidr_len > 32 || cidr_len < 8) {
		fprintf(stderr,
			"%s: valid CIDR prefix lengths are between 8 and 32\n",
			argv[0]);
		error++;
	    }
	    break;
	case 'a':
	    config.cnames = 1;
	    break;
	case 's':
	    config.stats = 1;
	    break;
	case 'b':
	    config.binary = 1;
	    break;
	case 'f':
//	    config.est_prec = 6;
	    config.fce_report_size = atoi(optarg);
	    config.fce = !!config.fce_report_size;
	    break;
	case 'y':
	    config.syn_counting = 1;
	    break;
	case 'q':
	    config.mem = 1;
	    break;
	case 'A':
	    config.anf_report_size = atoi(optarg);
	    config.anf = !!config.anf_report_size;
	    break;
	case 'S': // XXX
	    seeded++;
	    srandom(atoi(optarg)); // XXX
	    break;
	case 'h':
	    config.pretty = 1;
	    break;
	case 'r':
	    config.rotate_files = 1;
	    config.outfilename = rotating_filename;
	    break;
	case 'O':
	    config.rotate_files = 1;
	    /* fall through */
	case 'o':
	    config.outfilename = strdup(optarg);
	    break;
	case 'K':
	    config.pkt_prob = SampProb(atoi(optarg));
	    break;
#if 0
	case 'Y':
	    config.byte_prob = SampProb(atoi(optarg));
	    break;
#endif
	case 'F':
	    config.flow_prob = SampProb(atoi(optarg));
	    break;
	case 'e': // XXX does nothing --rkoga
//	    config.skip_ephemeral_ports = 1;
	    break;
	case 'n':
	    config.normalize_immediately = 1;
	    break;
	case 'u':
	    config.unscaled = 1;
	    break;
	case 'm':
	    config.merge_ifaces = 1;
	    break;
	case 'R':
	    config.fork = 1;
	    break;
	case '4':
	    config.allow_v4 = true;
	    break;
	case '6':
#ifdef HAVE_INET6
	    config.allow_v6 = true;
#else
	    fprintf(stderr, "IPv6 support is not enabled.\n");
#endif
	    break;
	default:
	    error++;
	    break;
	}
    }

    if (!config.allow_v4 && !config.allow_v6) {
	config.allow_v4 = true;
#ifdef HAVE_INET6
	config.allow_v6 = true;
#endif
    }

    if (error) {
	coral_usage(argv[0], "[<options>] <source>...\n"
"    Estimates bytes, packets, and (optionally) flows seen in network\n"
"    traffic, as described in \"Building a Better NetFlow\" [Estan, Keys,\n"
"    Moore 2004].  The Adaptive NetFlow (ANF) algorithm samples packets\n"
"    and generates tuple table entries with byte and packet counts.  The\n"
"    sampling rate adapts dynamically to keep the size of the tuple table\n"
"    within the desired range.  An entry's counters do not show the actual\n"
"    counts seen for that tuple, but rather are scaled to indicate the count\n"
"    of the set of tuples of which it is a sample.  There are two options for\n"
"    flow estimation.  The Flow Counting Extension (FCE) adaptively samples\n"
"    flows and generates additional tuple table entries with scaled flow\n"
"    counts.  SYN counting adds scaled flow counts into tuple table entries\n"
"    generated by ANF, but counts only TCP flows that start within the\n"
"    measurement interval, and is less accurate than FCE.\n"
	    "\n"
	    "Options:\n"
	    "-A<n>      report <n> table entries for ANF\n"
	    "-f<n>      count flows with FCE, reporting <n> table entries\n"
	    "           (to do FCE without ANF, use \"-f<n> -A0\")\n"
	    "-y         count flows with SYN counting\n"
            "-4         count IPv4 only\n"
            "-6         count IPv6 only\n"
            "-46        count IPv4 and IPv6 (default)\n"
	    "-p<n>      use <n> as length of CIDR prefix (default %d)\n"
	    "-a         resolve IP addresses to hostnames (only makes sense with -p32)\n"
	    "-b         print binary table data\n"
	    "-h         print human readable table data\n"
	    "-o<file>   specify the name of the output file (default: -)\n"
	    "-r         rotate the output file every interval (requires a filename with\n"
	    "           %% specifiers; the default filename with -r is '%s')\n"
	    "-O<file>   equivalent to -r -o<file>\n"
	    "-R         with -r, do reporting from a child process\n\n"
	    "-m         merge streams of multiple interfaces (and subinterfaces) into one\n"
	    "Performance parameters:\n"
#if 0
	    "-e         \"fold\" ephemeral ports\n" // XXX does nothing --rkoga
#endif
	    "-K<d>      set initial ANF sampling rate to 1/<d> packets\n" // XXX
#if 0
	    "-Y<d>      set initial ANF sampling rate to 1/<d> bytes\n" // XXX
#endif
	    "-F<d>      set initial FCE sampling rate to 1/<d> flows\n" // XXX
	    "Testing options (not for use in normal network measurement):\n"
	    "-n         normalize immediately instead of in parallel\n" // XXX
	    "-q         periodically print memory usage to stderr\n" // XXX
	    "-s         print hash table statistics\n"
	    "-S<n>      use random number seed <n>\n"
	    "-u         print unscaled counters in tuple table\n"
	    "\n"
"    <file> is the name of the output file, which may contain %% specifiers\n"
"    for formatting a timestamp.  For -o, the timestamp is the beginning of\n"
"    the trace; for -O, the timestamp is the beginning of the interval.\n"
"    The %% specifiers are any of those allowed by strftime(), plus:\n"
"        %%s   number of (whole) seconds since 1970-01-01 00:00:00 UTC\n"
"        %%f   fractional part of seconds (6 decimal places)\n"
"        %%F   equivalent to %%Y-%%m-%%d\n"
"    Except for %%f, all specifiers may contain printf-style modifiers\n"
"    (e.g., \"%%010s\").\n"
"    For -o, '-' means standard output.\n",
	    CIDR_LEN,
	    rotating_filename);
	exit(-1);
    }

    if (!config.rotate_files)
	config.fork = 0;
    if (config.anf && config.anf_report_size == 0) {
	coral_diag(0, ("%s: You must specify an ANF report size with -A<size>.\n", argv[0]));
	exit(-1);
    }
    config.anf_normal_size = config.anf_report_size * 2;
    config.anf_max_size = config.anf_report_size * 3;
    if (config.fce && config.syn_counting) {
	coral_diag(0, ("%s: FCE and SYN counting are incompatible.\n", argv[0]));
	exit(-1);
    }
    if (config.syn_counting && config.anf_report_size == 0) {
	coral_diag(0, ("%s: SYN counting requires a nonzero ANF report size.\n", argv[0]));
	exit(-1);
    }
    if (!seeded) srandom(time(NULL) ^ getpid()); /* XXX use /dev/random */

    if ((config.anon = coral_get_anonymizer())) {
	if (config.cnames) {
	    coral_diag(1, ("can not mix -a and -Canon options\n"));
	    exit(-1);
	}
	coral_set_anonymizer(NULL);
    }

    while (optind < argc) {
	if (!coral_new_source(argv[optind]))
	    exit(-1);
	optind++;
    }

    if (!coral_next_source(NULL)) {
	coral_diag(1, ("%s: warning: no sources specified.\n", argv[0]));
    }

    init_hashes();

    /* generate a netmask from the cidr length */
    config.netmask = 0xFFFFFFFF << (32 - cidr_len);

    /* corrected for endianess */
    config.netmask = ntohl(config.netmask);

    if (coral_open_all() <= 0)
	exit(-1);

    if (coral_start_all() < 0)
	exit(-1);

    signal(SIGINT, stop);

    coral_get_interval(&interval);
    config.dinterval = timevaltodouble(&interval);
    interval_data.begin = 0;
    interval_data.end = 0;
    interval_data.pkt_stats = NULL;

    duration = coral_get_duration();
    if (duration)
	coral_diag(2, ("collection duration max set to %d second(s)\n", 
		       duration));

    coral_read_pkt_init(NULL, NULL, &interval);

    p = file_header;
    p += sprintf(p, "# crl_anf output version: %s (%s format)\n",
	    OUTPUT_VERSION,
	    config.binary ? "binary" : config.pretty ? "pretty" : "text" );
    p += sprintf(p, "# generated by: ");
    for (i = 0; i < argc; i++)
	p += sprintf(p, " %s", argv[i]);
    sprintf(p, "\n");
    config.file_header = file_header;

    if (!config.binary) {
        if (!config.pretty) {
            tuple_key<in_addr>::fmt_lheader = "#%s\t%s\t";
            tuple_key<in_addr>::fmt_rheader = "%s\t%s\t%s\t%s\t";
            tuple_key<in_addr>::fmt_addr    = "%s\t";
            tuple_key<in_addr>::fmt_right   = "%u\t%u\t%u\t%u\t";
#ifdef HAVE_INET6
            tuple_key<in6_addr>::fmt_lheader = tuple_key<in_addr>::fmt_lheader;
            tuple_key<in6_addr>::fmt_rheader = tuple_key<in_addr>::fmt_rheader;
            tuple_key<in6_addr>::fmt_addr    = tuple_key<in_addr>::fmt_addr;
            tuple_key<in6_addr>::fmt_right   = tuple_key<in_addr>::fmt_right;
#endif
            fmt_dataheader = "%s\t%s\t%s\n";
            fmt_data = "%" PRIu64 "\t%" PRIu64 "\t%u\n";
        } else {
            if (config.cnames) {
                tuple_key<in_addr>::fmt_lheader = "#%-29s %-30s ";
                tuple_key<in_addr>::fmt_addr = "%-30s ";
            } else {
                tuple_key<in_addr>::fmt_lheader = "#%-14s %-15s ";
                tuple_key<in_addr>::fmt_addr = "%-15s ";
            }
            tuple_key<in_addr>::fmt_rheader = "%-3s %-2s %5s %5s ";
            tuple_key<in_addr>::fmt_right   = "%3u %2u %5u %5u ";
#ifdef HAVE_INET6
	    tuple_key<in6_addr>::fmt_lheader = "#%-38s %-39s ";
	    tuple_key<in6_addr>::fmt_addr    = "%-39s ";
            tuple_key<in6_addr>::fmt_rheader = tuple_key<in_addr>::fmt_rheader;
            tuple_key<in6_addr>::fmt_right   = tuple_key<in_addr>::fmt_right;
#endif
            fmt_dataheader = "%13s %15s %9s\n";
            fmt_data = "%13" PRIu64 " %15" PRIu64 " %9u\n";
        }
    }

    config.rf = coral_rf_open_file(&config.file, config.outfilename, "", 0);
    if (!config.rf)
	exit(1);

#ifndef NDEBUG
    procstats("start loop");
#endif

    while (1) {
	if (interrupted) {
	    coral_stop_all();
	    interrupted = 0;
	}
	iface = coral_read_pkt(&pkt_result, &interval_result);

	if (!iface) {
	    if (errno == EINTR || errno == EAGAIN) continue;
	    retval = errno;
	    break;
	}

	if (!pkt_result.packet && !interval_result.stats) {
	    /* beginning of interval */
#ifndef NDEBUG
	    procstats("start int");
#endif
	    double begin = timevaltodouble(&interval_result.begin);
	    interval_data.begin = begin;

	} else if (pkt_result.packet) {
	    /* got packet */
	    count_subif_stats(iface, &pkt_result);
	    if (config.mem) {
		static uint64_t n = 0;
		if (n++ % 10000 == 0) {
		    struct timespec ts;
		    coral_read_clock_timespec(iface, pkt_result.timestamp, &ts);
#ifndef NDEBUG
		    procstats.mem(&ts);
#endif
		}
	    }

	} else if (!pkt_result.packet && interval_result.stats) {
	    /* end of interval */
	    struct timeval tv1, tv2;
#ifndef NDEBUG
	    procstats("end int");
#endif
	    interval_data.end = timevaltodouble(&interval_result.end);
	    interval_data.pkt_stats = interval_result.stats;

	    gettimeofday(&tv1, NULL);

	    dump_data_fork(&interval_result.begin);

	    clear_data();
	    gettimeofday(&tv2, NULL);
	    timersub(&tv2, &tv1, &tv1);
	    coral_diag(2, ("crl_anf: closed %f s interval in %d.%06d s\n",
		interval_data.end - interval_data.begin,
		tv1.tv_sec, tv1.tv_usec));

	    /* reset the interval time, and other timers */
	    interval_data.begin = 0;
	    interval_data.end = 0;
	    interval_data.pkt_stats = NULL;
	}
    }

    coral_rf_close(config.rf);
    coral_stop_all();

    /* clean up stuff */
    if (config.stats) dump_hash_stats();
#if 0
    clear_hash_table(subif_hash);
    srcip_tab.clear();
    dstip_tab.clear();
#endif

    reap_children(0);

    return retval;
}

