/* 
 * Copyright 2008, 2009, 2011, 2012, 2019 The Regents of the University of California
 * All Rights Reserved 
 * 
 * Permission to use, copy, modify and distribute any part of this
 * CoralReef software package for educational, research and non-profit
 * purposes, without fee, and without a written agreement is hereby
 * granted, provided that the above copyright notice, this paragraph
 * and the following paragraphs appear in all copies.
 * 
 * Those desiring to incorporate this into commercial products or use
 * for commercial purposes should contact the Technology Transfer
 * Office, University of California, San Diego, 9500 Gilman Drive, La
 * Jolla, CA 92093-0910, Ph: (858) 534-5815, FAX: (858) 534-7345.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY
 * PARTY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL
 * DAMAGES, INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE SOFTWARE PROVIDED HEREIN IS ON AN "AS IS" BASIS, AND THE
 * UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE,
 * SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS. THE UNIVERSITY
 * OF CALIFORNIA MAKES NO REPRESENTATIONS AND EXTENDS NO WARRANTIES
 * OF ANY KIND, EITHER IMPLIED OR EXPRESS, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR A
 * PARTICULAR PURPOSE, OR THAT THE USE OF THE SOFTWARE WILL NOT INFRINGE
 * ANY PATENT, TRADEMARK OR OTHER RIGHTS.
 * 
 * The CoralReef software package is developed by the CoralReef
 * development team at the University of California, San Diego under
 * the Cooperative Association for Internet Data Analysis (CAIDA)
 * Program. Support for this effort is provided by the CAIDA grant
 * NCR-9711092, DARPA NGI Contract N66001-98-2-8922, DARPA NMS Grant
 * N66001-01-1-8909, and by CAIDA members.
 * 
 * Report bugs and suggestions to coral-bugs@caida.org.
 */

/*
 *
 * $Id: set_pcap_snaplen.c,v 1.5 2019/03/28 23:40:34 kkeys Exp $
 *
 */

#include "config.h"
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <fcntl.h>

#include <pcap.h>
#include "libcoral.h"
#include "crl_byteorder.h"

uint32_t snaplen = 0xffff;

static void usage(const char *name)
{
    fprintf(stderr, "Usage: %s [<options>] <pcapfile>...\n"
	"If <pcapfile> has a snaplen of 0, this app sets it to nonzero.\n"
	"File will be written in place, not copied.\n"
	"Options:\n"
	"-s<n>     set snaplen to <n> (default: %d)\n"
	"-f        force set snaplen even if it is not 0\n"
	"\n",
	name,
	snaplen);
}

int main(int argc, char *argv[])
{
    int force = 0;
    int swapped = 0;
    struct pcap_file_header pfh;
    int fd;
    ssize_t len;
    int success = 0;
    int errors = 0;
    int opt;

    while ((opt = getopt(argc, argv, "s:f")) >= 0) {
	switch (opt) {
	case 's':
	    snaplen = atoi(optarg);  break;
	case 'f':
	    force = 1; break;
        default:
            usage(argv[0]);
            exit(-1);
        }
    }

    for ( ; optind < argc; optind++) {
	fd = open(argv[optind], O_RDWR);
	if (fd < 0) {
	    fprintf(stderr, "can't open %s: %s\n",
		argv[optind], strerror(errno));
	    errors++; continue;
	}

	len = read(fd, (void*)&pfh, sizeof(pfh));
	if (len < 0) {
	    fprintf(stderr, "can't read %s: %s\n",
		argv[optind], strerror(errno));
	    errors++; close(fd); continue;
	}
	if (len < sizeof(pfh)) {
	    fprintf(stderr, "%s is not a valid pcap file\n", argv[optind]);
	    errors++; close(fd); continue;
	}
	if (pfh.magic == 0xa1b2c3d4) {
	    swapped = 0;
	} else if (pfh.magic == 0xd4c3b2a1) {
	    swapped = 1;
	} else {
	    fprintf(stderr, "%s is not a pcap file (magic=%08x)\n",
		argv[optind], pfh.magic);
	    errors++; close(fd); continue;
	}
	if (pfh.snaplen != 0 && !force) {
	    fprintf(stderr, "%s has nonzero snaplen %d\n", argv[optind],
		swapped ? crl_swapl(pfh.snaplen) : pfh.snaplen);
	    errors++; close(fd); continue;
	}
	pfh.snaplen = swapped ? crl_swapl(snaplen) : snaplen;
	if (lseek(fd, SEEK_SET, 0) < 0) {
	    fprintf(stderr, "can't seek %s: %s\n",
		argv[optind], strerror(errno));
	    errors++; close(fd); continue;
	}
	len = write(fd, (void*)&pfh, sizeof(pfh));
	if (len < 0) {
	    fprintf(stderr, "can't write %s: %s\n",
		argv[optind], strerror(errno));
	    errors++; close(fd); continue;
	}
	success++;
	close(fd);
    }

    fprintf(stderr, "%d files changed, %d errors\n", success, errors);

    exit(errors);
}
