#!/usr/bin/env perl
## 
## Copyright 2011, 2012, 2019 The Regents of the University of California
## All Rights Reserved 
## 
## Permission to use, copy, modify and distribute any part of this
## CoralReef software package for educational, research and non-profit
## purposes, without fee, and without a written agreement is hereby
## granted, provided that the above copyright notice, this paragraph
## and the following paragraphs appear in all copies.
## 
## Those desiring to incorporate this into commercial products or use
## for commercial purposes should contact the Technology Transfer
## Office, University of California, San Diego, 9500 Gilman Drive, La
## Jolla, CA 92093-0910, Ph: (858) 534-5815, FAX: (858) 534-7345.
## 
## IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY
## PARTY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL
## DAMAGES, INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS
## SOFTWARE, EVEN IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF
## THE POSSIBILITY OF SUCH DAMAGE.
## 
## THE SOFTWARE PROVIDED HEREIN IS ON AN "AS IS" BASIS, AND THE
## UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE,
## SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS. THE UNIVERSITY
## OF CALIFORNIA MAKES NO REPRESENTATIONS AND EXTENDS NO WARRANTIES
## OF ANY KIND, EITHER IMPLIED OR EXPRESS, INCLUDING, BUT NOT LIMITED
## TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR A
## PARTICULAR PURPOSE, OR THAT THE USE OF THE SOFTWARE WILL NOT INFRINGE
## ANY PATENT, TRADEMARK OR OTHER RIGHTS.
## 
## The CoralReef software package is developed by the CoralReef
## development team at the University of California, San Diego under
## the Cooperative Association for Internet Data Analysis (CAIDA)
## Program. Support for this effort is provided by the CAIDA grant
## NCR-9711092, DARPA NGI Contract N66001-98-2-8922, DARPA NMS Grant
## N66001-01-1-8909, and by CAIDA members.
## 
## Report bugs and suggestions to coral-bugs@caida.org.
## 
## Parses output of (only!) flowd-reader -Uc
## Not a general use script, but can be modified for other netflow outputs.

use strict;
use warnings;

my $coral_dir;
BEGIN { $coral_dir = "/usr/local/Coral"; } # This will be edited during build.
use lib "$coral_dir/lib"; #coral global libdir#
use lib "../../lib"; #coral local libdir#
use CAIDA::Traffic2::FileWriter;
use CAIDA::Tables::Tuple_Table;

$| = 1;                 # don't buffer output

my ($pname) = $0 =~ m#([^/]+)$#;

my $writer = new CAIDA::Traffic2::FileWriter(*STDOUT);
my $tuple_table = new CAIDA::Tables::Tuple_Table;
my $interval = new CAIDA::Traffic2::Interval;
my $id_info = new CAIDA::Traffic2::SubinterfaceInfo;
my $id = 0;

$id_info->_set_table('Tuple Table', $tuple_table);
$id_info->set_metadata('id', $id);
$id_info->set_metadata('v4 packets', 50);
$id_info->set_metadata('save tables', 1);
$interval->set_metadata('ids', [$id]);
$interval->{'id info'}{$id} = $id_info;
$writer->initialize($interval, 'Tuple Table');

my $interval_size = 300;
my $interval_start;
my $interval_end;

my ($tot_bytes, $tot_pkts, $tot_flows) = (0,0,0);

while (<>) {
    if (/^#:(.*)/) {
	my @field_names = split /,/, $1;
	# Sanity check on field order:
	if ($field_names[0] ne 'unix_secs' or $field_names[4] ne 'dpkts' or
	    $field_names[6] ne 'first' or $field_names[17] ne 'prot')
	{
	    die "Unexpected csv order!";
	}
    } else {
	my ($unix_secs, $unix_nsecs, $sysuptime, $exaddr, $dpkts, $doctets, $first, $last, $engine_type, $engine_id, $srcaddr, $dstaddr, $nexthop, $input, $output, $srcport, $dstport, $prot, $tos, $tcp_flags, $src_mask, $dst_mask, $src_as, $dst_as) = split /,/;
	my $recv_time = $unix_secs + $unix_nsecs/1e9;
	if (not defined $interval_start) {
	    $interval_start = $recv_time;
	    $interval_end = int($interval_start/$interval_size + 1) * $interval_size;
	}
	if ($recv_time > $interval_end) {
	    #Spit out old interval, create new one.
	    $id_info->set_metadata('v4 bytes', $tot_bytes);
	    $id_info->set_metadata('v4 packets', $tot_pkts);
	    $id_info->set_metadata('v4 flows', $tot_flows);
	    $interval->set_metadata('start', $interval_start);
	    $interval->set_metadata('duration', $interval_end-$interval_start);
	    $interval->set_metadata('end', $interval_end);
	    $writer->dump_interval_full();
	    $interval_start = $interval_end;
	    # NB: Makes the assumption of constant data flow, without big
	    # gaps in collection.
	    $interval_end += $interval_size;
	    $tuple_table = new CAIDA::Tables::Tuple_Table;
	    $id_info->_set_table('Tuple Table', $tuple_table);
	    ($tot_bytes, $tot_pkts, $tot_flows) = (0,0,0);
	}
	my $first_delta = $sysuptime - $first;
	my $last_delta = $sysuptime - $last;
	$tot_bytes += $doctets;
	$tot_pkts += $dpkts;
	$tot_flows++;
	my $counts = new CAIDA::Traffic2::FlowCounter($dpkts, $doctets, 1,
		$recv_time - $first_delta/1e3, $recv_time - $last_delta/1e3);
	$tuple_table->entry_add($srcaddr, $dstaddr, $prot, 1,
				$srcport, $dstport, $counts);
    }
}

if (not defined $interval_start) { # No data read at all
    print STDERR "No data found, trying to fail gracefully.\n";
    $interval_start = $interval_end = 0;
}

# Print final interval.
$id_info->set_metadata('v4 bytes', $tot_bytes);
$id_info->set_metadata('v4 packets', $tot_pkts);
$id_info->set_metadata('v4 flows', $tot_flows);
$interval->set_metadata('start', $interval_start);
$interval->set_metadata('duration', $interval_end-$interval_start);
$interval->set_metadata('end', $interval_end);
$writer->dump_interval_full();
