/* 
 * Copyright 2008, 2011, 2012, 2019
 * The Regents of the University of California 
 * All Rights Reserved 
 * 
 * Permission to use, copy, modify and distribute any part of this
 * CoralReef software package for educational, research and non-profit
 * purposes, without fee, and without a written agreement is hereby
 * granted, provided that the above copyright notice, this paragraph
 * and the following paragraphs appear in all copies.
 * 
 * Those desiring to incorporate this into commercial products or use
 * for commercial purposes should contact the Technology Transfer
 * Office, University of California, San Diego, 9500 Gilman Drive, La
 * Jolla, CA 92093-0910, Ph: (858) 534-5815, FAX: (858) 534-7345.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY
 * PARTY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL
 * DAMAGES, INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE SOFTWARE PROVIDED HEREIN IS ON AN "AS IS" BASIS, AND THE
 * UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE,
 * SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS. THE UNIVERSITY
 * OF CALIFORNIA MAKES NO REPRESENTATIONS AND EXTENDS NO WARRANTIES
 * OF ANY KIND, EITHER IMPLIED OR EXPRESS, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR A
 * PARTICULAR PURPOSE, OR THAT THE USE OF THE SOFTWARE WILL NOT INFRINGE
 * ANY PATENT, TRADEMARK OR OTHER RIGHTS.
 * 
 * The CoralReef software package is developed by the CoralReef
 * development team at the University of California, San Diego under
 * the Cooperative Association for Internet Data Analysis (CAIDA)
 * Program.  Support for this effort is provided by the CAIDA grant
 * NCR-9711092, DARPA NGI Contract N66001-98-2-8922, DARPA NMS Grant
 * N66001-01-1-8909, and by CAIDA members.
 * 
 * Report bugs and suggestions to coral-bugs@caida.org.
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "libcoral.h"
#include "ipset.h"

typedef struct {
    unsigned char bitmap[1<<(16-3)];
} ipset_node16_t;

typedef struct {
    ipset_node16_t *node16[1<<8];
} ipset_node8_t;

struct ipset {
    ipset_node8_t *node8[1<<8];
};

static uint32_t ipset_bitcount[] = {
    0, 1, 1, 2, 1, 2, 2, 3, 1, 2, 2, 3, 2, 3, 3, 4,
    1, 2, 2, 3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 5,
    1, 2, 2, 3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 5,
    2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6,
    1, 2, 2, 3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 5,
    2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6,
    2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6,
    3, 4, 4, 5, 4, 5, 5, 6, 4, 5, 5, 6, 5, 6, 6, 7,
    1, 2, 2, 3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 5,
    2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6,
    2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6,
    3, 4, 4, 5, 4, 5, 5, 6, 4, 5, 5, 6, 5, 6, 6, 7,
    2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6,
    3, 4, 4, 5, 4, 5, 5, 6, 4, 5, 5, 6, 5, 6, 6, 7,
    3, 4, 4, 5, 4, 5, 5, 6, 4, 5, 5, 6, 5, 6, 6, 7,
    4, 5, 5, 6, 5, 6, 6, 7, 5, 6, 6, 7, 6, 7, 7, 8,
};

ipset_t *ipset_new() {
    return calloc(1, sizeof(ipset_t));
}

void ipset_add(ipset_t *ipset, uint32_t a) {
    unsigned part1 = a>>24;
    unsigned part2 = (a>>16) & 0xFF;
    unsigned part3 = (a & 0xFFFF) >> 3;
    unsigned part4 = a & 0x7;
    if (!ipset->node8[part1]) {
	ipset->node8[part1] = calloc(1, sizeof(ipset_node8_t));
    }
    if (!ipset->node8[part1]->node16[part2]) {
	ipset->node8[part1]->node16[part2] = calloc(1, sizeof(ipset_node16_t));
    }
    ipset->node8[part1]->node16[part2]->bitmap[part3] |= (1 << part4);
}

int ipset_test(ipset_t *ipset, uint32_t a) {
    unsigned part1 = a>>24;
    unsigned part2 = (a>>16) & 0xFF;
    unsigned part3 = (a & 0xFFFF) >> 3;
    unsigned part4 = a & 0x7;
    if (!ipset->node8[part1])
	return 0;
    if (!ipset->node8[part1]->node16[part2])
	return 0;
    return !!(ipset->node8[part1]->node16[part2]->bitmap[part3] & (1 << part4));
}

uint32_t ipset_count(ipset_t *ipset) {
    uint32_t count = 0;
    int i, j, k;
    for (i = 0; i < (1<<8); i++) {
	if (ipset->node8[i]) {
	    for (j = 0; j < (1<<8); j++) {
		if (ipset->node8[i]->node16[j]) {
		    for (k = 0; k < (1<<(16-3)); k++) {
			count +=
			    ipset_bitcount[ipset->node8[i]->node16[j]->bitmap[k]];
		    }
		}
	    }
	}
    }
    return count;
}

void ipset_dump(ipset_t *ipset) {
    int i, j, k;
    struct in_addr addr;
    for (i = 0; i < (1<<8); i++) {
	if (ipset->node8[i]) {
	    for (j = 0; j < (1<<8); j++) {
		if (ipset->node8[i]->node16[j]) {
		    for (k = 0; k < (1<<16); k++) {
			if (ipset->node8[i]->node16[j]->bitmap[k>>3] & (1<<(k&7))) {
			    addr.s_addr = (i << 24) | (j << 16) | k;
			    addr.s_addr = htonl(addr.s_addr);
			    printf("%s\n", inet_ntoa(addr));
			}
		    }
		}
	    }
	}
    }
}
