#ifndef IPV4ADDR_HH
#define IPV4ADDR_HH

extern "C" {
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
}

#include <exception>
#include <stdexcept>

class ipv4addr_t : private in_addr {
    ipv4addr_t(const uint32_t &a) { s_addr = a; }
public:
    static const int family = AF_INET;
    ipv4addr_t() { s_addr = htonl(0); }
    ipv4addr_t(const in_addr &src) { s_addr = src.s_addr; }
    ipv4addr_t(const char *str) { parse(str); }
    void parse(const char *str) {
	if (inet_pton(AF_INET, str, this) != 1) {
	    throw std::runtime_error(std::string("can't parse IPv4 address ") + str);
	}
    }

    ipv4addr_t operator= (const in_addr &src)
	{ s_addr = src.s_addr; return *this; }

    uint32_t to_hl() const { return ntohl(s_addr); }
    uint32_t to_nl() const { return s_addr; }

    // not implemented in order to avoid accidents.
    //operator uint32_t() const { return ntohl(s_addr); }

    bool operator < (const ipv4addr_t &b) const
	{ return ntohl(s_addr) < ntohl(b.s_addr); }

    bool operator == (const ipv4addr_t &b) const
	{ return s_addr == b.s_addr; }

#if 0 // untested, not needed (yet)
    ipv4addr_t operator + (const uint32_t &b)
	{ return ipv4addr_t(htonl(ntohl(s_addr) + b)); }
    ipv4addr_t operator - (const uint32_t &b)
	{ return ipv4addr_t(htonl(ntohl(s_addr) - b)); }
    ipv4addr_t operator ++ ();
    ipv4addr_t operator ++ (int bogus);
    ipv4addr_t operator -- ();
    ipv4addr_t operator -- (int bogus);
#endif

    ipv4addr_t operator & (const ipv4addr_t &b) const
	{ return ipv4addr_t(s_addr & b.s_addr); }
    ipv4addr_t operator | (const ipv4addr_t &b) const
	{ return ipv4addr_t(s_addr | b.s_addr); }
    ipv4addr_t operator ^ (const ipv4addr_t &b) const
	{ return ipv4addr_t(s_addr ^ b.s_addr); }

    ipv4addr_t & operator &= (const ipv4addr_t &b)
	{ s_addr &= b.s_addr; return *this; }
    ipv4addr_t & operator |= (const ipv4addr_t &b)
	{ s_addr |= b.s_addr; return *this; }
    ipv4addr_t & operator ^= (const ipv4addr_t &b)
	{ s_addr ^= b.s_addr; return *this; }

    bool bittest(size_t k) const {
	return ((const uint8_t*)(this))[k/8] & (1 << ((31-k)%8));
    }

    static ipv4addr_t netmask(const int &len) {
	if (len == 0) return ipv4addr_t();
	return ipv4addr_t(htonl(0xffffffff << (32 - len)));
    }

    friend std::ostream & operator << (std::ostream & os, const ipv4addr_t &a)
	{ return os << inet_ntoa(a); }
};

#endif // IPV4ADDR_HH
