#ifndef IPV4HADDR_HH
#define IPV4HADDR_HH

extern "C" {
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
}

#include <exception>
#include <stdexcept>

class ipv4haddr_t {
    uint32_t haddr;
    ipv4haddr_t(uint32_t a) { haddr = a; }
public:
    static const int family = AF_INET;
    ipv4haddr_t() { haddr = 0; }
    ipv4haddr_t(const in_addr &src) { haddr = ntohl(src.s_addr); }
    void parse(const char *str) {
	in_addr naddr;
	if (inet_pton(AF_INET, str, &naddr) != 1) {
	    throw std::runtime_error(std::string("can't parse IPv4 address ") + str);
	}
	haddr = ntohl(naddr.s_addr);
    }

    ipv4haddr_t operator= (const in_addr &src)
	{ haddr = ntohl(src.s_addr); return *this; }

    uint32_t to_hl() const { return haddr; }
    uint32_t to_nl() const { return htonl(haddr); }

    // not implemented in order to avoid accidents.
    //operator uint32_t() const { return haddr; }

    bool operator < (const ipv4haddr_t &b) const
	{ return haddr < b.haddr; }

    bool operator == (const ipv4haddr_t &b) const
	{ return haddr == b.haddr; }

#if 0 // untested, not needed (yet)
    ipv4haddr_t operator + (const uint32_t &b)
	{ return ipv4haddr_t(haddr + b); }
    ipv4haddr_t operator - (const uint32_t &b)
	{ return ipv4haddr_t(haddr - b); }
    ipv4haddr_t operator ++ ();
    ipv4haddr_t operator ++ (int bogus);
    ipv4haddr_t operator -- ();
    ipv4haddr_t operator -- (int bogus);
#endif

    ipv4haddr_t operator & (const ipv4haddr_t &b) const
	{ return ipv4haddr_t(haddr & b.haddr); }
    ipv4haddr_t operator | (const ipv4haddr_t &b) const
	{ return ipv4haddr_t(haddr | b.haddr); }
    ipv4haddr_t operator ^ (const ipv4haddr_t &b) const
	{ return ipv4haddr_t(haddr ^ b.haddr); }

    ipv4haddr_t & operator &= (const ipv4haddr_t &b)
	{ haddr &= b.haddr; return *this; }
    ipv4haddr_t & operator |= (const ipv4haddr_t &b)
	{ haddr |= b.haddr; return *this; }
    ipv4haddr_t & operator ^= (const ipv4haddr_t &b)
	{ haddr ^= b.haddr; return *this; }

    bool bittest(size_t k) const {
	return haddr & (1 << (31-k));
    }

    static ipv4haddr_t netmask(const int &len) {
	if (len == 0) return ipv4haddr_t(0);
	return ipv4haddr_t(0xffffffff << (32 - len));
    }

    friend std::ostream & operator << (std::ostream & os, const ipv4haddr_t &a)
    {
	in_addr naddr;
	naddr.s_addr = htonl(a.haddr);
	return os << inet_ntoa(naddr);
    }
};

#endif // IPV4HADDR_HH
