## 
## Copyright 2000, 2001, 2003, 2004, 2005, 2006, 2007, 2008, 2009, 2011, 2012, 2019 The Regents of the University of California
## All Rights Reserved 
## 
## Permission to use, copy, modify and distribute any part of this
## CoralReef software package for educational, research and non-profit
## purposes, without fee, and without a written agreement is hereby
## granted, provided that the above copyright notice, this paragraph
## and the following paragraphs appear in all copies.
## 
## Those desiring to incorporate this into commercial products or use
## for commercial purposes should contact the Technology Transfer
## Office, University of California, San Diego, 9500 Gilman Drive, La
## Jolla, CA 92093-0910, Ph: (858) 534-5815, FAX: (858) 534-7345.
## 
## IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY
## PARTY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL
## DAMAGES, INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS
## SOFTWARE, EVEN IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF
## THE POSSIBILITY OF SUCH DAMAGE.
## 
## THE SOFTWARE PROVIDED HEREIN IS ON AN "AS IS" BASIS, AND THE
## UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE,
## SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS. THE UNIVERSITY
## OF CALIFORNIA MAKES NO REPRESENTATIONS AND EXTENDS NO WARRANTIES
## OF ANY KIND, EITHER IMPLIED OR EXPRESS, INCLUDING, BUT NOT LIMITED
## TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR A
## PARTICULAR PURPOSE, OR THAT THE USE OF THE SOFTWARE WILL NOT INFRINGE
## ANY PATENT, TRADEMARK OR OTHER RIGHTS.
## 
## The CoralReef software package is developed by the CoralReef
## development team at the University of California, San Diego under
## the Cooperative Association for Internet Data Analysis (CAIDA)
## Program. Support for this effort is provided by the CAIDA grant
## NCR-9711092, DARPA NGI Contract N66001-98-2-8922, DARPA NMS Grant
## N66001-01-1-8909, and by CAIDA members.
## 
## Report bugs and suggestions to coral-bugs@caida.org.
## 

package CAIDA::Traffic2::FileReader;

use strict;
use Carp;
require CAIDA::Traffic2::Interval;

my $cvs_Id = '$Id: FileReader.pm,v 1.32 2019/03/28 23:40:56 kkeys Exp $';
my $cvs_Author = '$Author: kkeys $';
my $cvs_Name = '$Name: release-3-9-5 $';
my $cvs_Revision = '$Revision: 1.32 $';

my $min_version = "1.1";

my $version_pat = 'version: (.*)$';
my $args_pat = 'generated by:  (.*)$';

sub new {
    my ($classname, $file) = @_;
    my $filehandle;
    if (not $file) {
	carp "FileReader requires a file name or file handle to read from.";
	return undef;
    } elsif (defined fileno $file) {
	$filehandle = $file;
    } elsif (ref $file) {
	carp "FileReader requires a file name or file handle to read from.";
	return undef;
    } else {
	require IO::File;
	my $new_handle = new IO::File "<$file";
	if (not $new_handle) {
	    carp "Cannot open file $file: $!";
	    return undef;
	}
	$filehandle = $new_handle;
    }
    my $self =  {
		    'file' => $filehandle,
		    'version' => undef,
		    'command line' => undef,
		    'conv options' => {},
		    'desired tables' => {},
		};
    bless $self;
    if ($] >= 5.008) {
	# Work around synchronization issues with PerlIO conversion to FILE*.
	# Only affects pipes.  Pushing :stdio layer fails to change anything.
	my $position = tell $filehandle;
	my $using_perlio = 0;
	my @layers;
	eval { @layers = PerlIO::get_layers($filehandle); };
	if ($@) {
	    carp "Cannot get filehandle layers in FileReader; " .
		    "may cause read errors.";
	} else {
	    foreach my $layer (@layers) {
		$using_perlio = 1 if $layer eq 'perlio';
	    }
	    if ($using_perlio) {
		if ($position == 0) {
		    # Only push the layer if nothing's been read.
		    # XXX This does NOT guarantee that no buffering has
		    # occurred, but in practice it seems to be true.
		    binmode($filehandle, ":unix");
		} else {
		    carp "Cannot set filehandle layer in FileReader;" .
			    "may cause read errors.";
		}
	    }
	}
    }
    my $header_ok = $self->_read_header($filehandle);
    unless ($header_ok) {
	carp "Unknown format in FileReader.";
	return undef;
    }
    return $self;
}

sub _read_header {
    my ($self, $filehandle) = @_;
    my $got_header = 0;
    while (<$filehandle>) {
	last if /^\s*$/;
	chomp;
	if (/$version_pat/o) {
	    $self->{'version'} = $1;
	    $got_header = 1;
	    if ($1 =~ /pretty/) {
		carp "CAIDA::Tables objects cannot handle human-readable " . 
		    "(pretty) output format.";
		return 0;
	    }
	    (my $file_ver = $1) =~ s/ \(.*\)//g;
	    if ($file_ver < $min_version) {
		carp "File format must be version $min_version or greater.";
		return 0;
	    }
	} elsif (/$args_pat/o) {
	    $self->{'command line'} = $1;
	}
    }
    return $got_header;
}

sub set_conversion_options {
    my ($self, $hashref) = @_;
    $self->{'conv options'} = $hashref;
}

sub set_desired_table_types {
    my ($self, @types) = @_;
    $self->{'desired tables'} = {}; # In case of a reset, clear out old vals.
    foreach my $type (@types) {
	$type =~ s/_/ /g;
	$self->{'desired tables'}{$type} = 1;
    }
}

sub get_next_interval {
    my ($self, $preload) = @_;
    my $filehandle = $self->{'file'};
    my ($interval, $interval_okay);
    unless (eof $filehandle) {
	$interval = new CAIDA::Traffic2::Interval;
	$interval_okay = $interval->_initialize($filehandle,
				    $self->{'desired tables'},
				    $self->{'conv options'}, $preload);
    }
    return undef if not $interval_okay;
    if ($preload) {
	foreach my $id_info ($interval->get_id_infos()) {
	    foreach my $type ($id_info->get_table_types()) {
		$interval->get_table($id_info, $type);
	    }
	}
    }
    return $interval;
}

sub get_metadata {
    my ($self, $field) = @_;
    if (not exists $self->{$field}) {
	carp "Unknown metadata field '$field'";
	return undef;
    }
    return $self->{$field};
}

1;
