/* 
 * Copyright 1999, 2000, 2001, 2003, 2004, 2005, 2006, 2007, 2008, 2009, 2011, 2012, 2019 The Regents of the University of California
 * All Rights Reserved 
 * 
 * Permission to use, copy, modify and distribute any part of this
 * CoralReef software package for educational, research and non-profit
 * purposes, without fee, and without a written agreement is hereby
 * granted, provided that the above copyright notice, this paragraph
 * and the following paragraphs appear in all copies.
 * 
 * Those desiring to incorporate this into commercial products or use
 * for commercial purposes should contact the Technology Transfer
 * Office, University of California, San Diego, 9500 Gilman Drive, La
 * Jolla, CA 92093-0910, Ph: (858) 534-5815, FAX: (858) 534-7345.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY
 * PARTY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL
 * DAMAGES, INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE SOFTWARE PROVIDED HEREIN IS ON AN "AS IS" BASIS, AND THE
 * UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE,
 * SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS. THE UNIVERSITY
 * OF CALIFORNIA MAKES NO REPRESENTATIONS AND EXTENDS NO WARRANTIES
 * OF ANY KIND, EITHER IMPLIED OR EXPRESS, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR A
 * PARTICULAR PURPOSE, OR THAT THE USE OF THE SOFTWARE WILL NOT INFRINGE
 * ANY PATENT, TRADEMARK OR OTHER RIGHTS.
 * 
 * The CoralReef software package is developed by the CoralReef
 * development team at the University of California, San Diego under
 * the Cooperative Association for Internet Data Analysis (CAIDA)
 * Program. Support for this effort is provided by the CAIDA grant
 * NCR-9711092, DARPA NGI Contract N66001-98-2-8922, DARPA NMS Grant
 * N66001-01-1-8909, and by CAIDA members.
 * 
 * Report bugs and suggestions to coral-bugs@caida.org.
 */

/* High-level functions for pcap interfaces (devices or tracefiles).
 *
 *  $Id: coral_iface_pcap.c,v 1.41 2019/03/28 23:41:00 kkeys Exp $
 *
 */

#include "config.h"
#include "coraldefs.h"
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/param.h>

#include "libcoral.h"
#include "libcoral_priv.h"
#include "coral_clock.h"

static const char RCSid[] = "$Id: coral_iface_pcap.c,v 1.41 2019/03/28 23:41:00 kkeys Exp $";


static inline void pcap_clock_order(const coral_iface_t *iface,
    const coral_timestamp_t *src, coral_timestamp_t *dst)
{
    *dst = *src;
}

static void inline pcap_clock_timespec(const coral_iface_t *iface,
    const coral_timestamp_t *t, struct timespec *tspec)
{
    tspec->tv_sec  = t->i[0];
    tspec->tv_nsec  = t->i[1] * 1000;
}

void coral_pcap_update_pkt_stats(coral_iface_t *iface)
{
    struct pcap_stat stats;

    if (iface->src->type.coral_type == CORAL_TYPE_PCAP)
	return; /* stats are not available for files */
    if (iface->eof)
	return; /* stats were updated by coral_pcaplive_stop() */

    if (pcap_stats(iface->src->file, &stats) != 0) {
	coral_diag(0, ("%s: %s\n", iface->src->filename,
	    pcap_geterr(iface->src->file)));
	return;
    }

    iface->pkt_stats.pkts_recv = stats.ps_recv - iface->u.pcap.stats.ps_recv;
    iface->pkt_stats.pkts_drop = stats.ps_drop - iface->u.pcap.stats.ps_drop;
    iface->u.pcap.stats = stats;
}

static void coral_pcap_report_pkt_drops(coral_iface_t *iface,
    struct timeval *when)
{
    struct pcap_stat stats;
    int64_t pkts_drop;
    if (pcap_stats(iface->src->file, &stats) != 0) {
	coral_diag(0, ("%s: %s\n", iface->src->filename,
	    pcap_geterr(iface->src->file)));
	return;
    }
    pkts_drop = stats.ps_drop - iface->old_pkts_drop;
    iface->old_pkts_drop = stats.ps_drop;
    if (pkts_drop) {
	coral_print_pkt_drops(iface, when, pkts_drop);
    }
}

static void pcap_iface_init(coral_iface_t *iface)
{
    iface->time_is_normal = 1;
}

static int pcap_consume(coral_iface_t *iface, int want)
{
    /* The read_min will block 1ms if there's no data, but consume is called
     * after a successful read, so there probably is more data. */
    return coral_read_any(iface) && iface->have_data;
}

const coral_iface_type_t coral_iface_type_pcap = {
    { -1, -1, -1, -1, -1, -1, -1, -1 },	/* pcap doesn't have ATM cells */
    "pcap",
    CORAL_API_PKT,
    pcap_iface_init,
    pcap_clock_order,
    NULL /* clock_read_double */,
    pcap_clock_timespec,
    NULL /* correct_clock */,
    NULL /* clock_native */,
    pcap_consume,
    NULL /* consume_pkt */,
    NULL /* prep_pkt */,
    coral_pcap_update_pkt_stats,
    coral_pcap_report_pkt_drops,
    NULL /* close */,
    { { CORAL_RX, 68 }, 0, CORAL_PROTO_UNKNOWN, CORAL_PROTO_UNKNOWN,
	NULL }			/* used by pcaplive only */
};

